# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only


# Handles the jingle signalling protocol


# TODO:
# * things in XEP 0176, including:
#      - http://xmpp.org/extensions/xep-0176.html#protocol-restarts
#      - http://xmpp.org/extensions/xep-0176.html#fallback
# * XEP 0177 (raw udp)

# * UI:
#   - make state and codec information available to the user
#   - video integration
#   * config:
#     - codecs

from __future__ import annotations

from typing import Any

import logging

import nbxmpp
from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Iq
from nbxmpp.structs import IqProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import types
from gajim.common.file_props import FileProp
from gajim.common.jingle_ft import JingleFileTransfer
from gajim.common.jingle_rtp import JingleAudio
from gajim.common.jingle_rtp import JingleVideo
from gajim.common.jingle_session import JingleSession
from gajim.common.jingle_session import JingleStates
from gajim.common.jingle_transport import JingleTransportIBB
from gajim.common.jingle_transport import JingleTransportSocks5
from gajim.common.modules.base import BaseModule
from gajim.common.util.jid import get_full_jid_from_iq
from gajim.common.util.jid import InvalidFormat
from gajim.common.util.text import get_random_string

logger = logging.getLogger("gajim.c.m.jingle")


class Jingle(BaseModule):
    def __init__(self, con: types.Client) -> None:
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(name="iq", typ="result", callback=self._on_jingle_iq),
            StanzaHandler(name="iq", typ="error", callback=self._on_jingle_iq),
            StanzaHandler(
                name="iq", typ="set", ns=Namespace.JINGLE, callback=self._on_jingle_iq
            ),
        ]

        # dictionary: sessionid => JingleSession object
        self._sessions: dict[str, JingleSession] = {}

        # dictionary: (jid, iq stanza id) => JingleSession object,
        # one time callbacks
        self.__iq_responses: dict[tuple[str, str], JingleSession] = {}
        self.files: list[dict[str, Any]] = []

    def delete_jingle_session(self, sid: str) -> None:
        """
        Remove a jingle session from a jingle stanza dispatcher
        """
        if sid in self._sessions:
            # FIXME: Move this elsewhere?
            for content in list(self._sessions[sid].contents.values()):
                content.destroy()
            self._sessions[sid].callbacks = []
            del self._sessions[sid]

    def _on_jingle_iq(
        self, _con: types.NBXMPPClient, stanza: Iq, _properties: IqProperties
    ) -> None:
        """
        The jingle stanza dispatcher

        Route jingle stanza to proper JingleSession object, or create one if it
        is a new session.

        TODO: Also check if the stanza isn't an error stanza, if so route it
        adequately.
        """
        # get data
        try:
            jid = get_full_jid_from_iq(stanza)
        except InvalidFormat:
            logger.warning("Invalid JID: %s, ignoring it", stanza.getFrom())
            return
        id_ = stanza.getID()
        if (jid, id_) in self.__iq_responses:
            self.__iq_responses[(jid, id_)].on_stanza(stanza)
            del self.__iq_responses[(jid, id_)]
            raise nbxmpp.NodeProcessed
        jingle = stanza.getTag("jingle")
        # a jingle element is not necessary in iq-result stanza
        # don't check for that
        if jingle:
            sid = jingle.getAttr("sid")
        else:
            sid = None
            for sesn in self._sessions.values():
                if id_ in sesn.iq_ids:
                    sesn.on_stanza(stanza)
            return

        # do we need to create a new jingle object
        if sid not in self._sessions:
            # TODO: tie-breaking and other things...
            newjingle = JingleSession(
                self._con, weinitiate=False, jid=jid, iq_id=id_, sid=sid
            )
            self._sessions[sid] = newjingle
        # we already have such session in dispatcher...
        self._sessions[sid].collect_iq_id(id_)
        self._sessions[sid].on_stanza(stanza)
        # Delete invalid/unneeded sessions
        if sid in self._sessions and self._sessions[sid].state == JingleStates.ENDED:
            self.delete_jingle_session(sid)
        raise nbxmpp.NodeProcessed

    def start_audio(self, jid: str) -> str:
        audio_session = self.get_jingle_session(jid, media="audio")
        if audio_session is not None:
            return audio_session.sid
        jingle = self.get_jingle_session(jid, media="video")
        if jingle:
            jingle.add_content("voice", JingleAudio(jingle))
        else:
            jingle = JingleSession(self._con, weinitiate=True, jid=jid)
            self._sessions[jingle.sid] = jingle
            jingle.add_content("voice", JingleAudio(jingle))
            jingle.start_session()
        return jingle.sid

    def start_video(self, jid: str) -> str:
        video_session = self.get_jingle_session(jid, media="video")
        if video_session is not None:
            return video_session.sid
        jingle = self.get_jingle_session(jid, media="audio")
        if jingle:
            video = JingleVideo(jingle)
            jingle.add_content("video", video)
        else:
            jingle = JingleSession(self._con, weinitiate=True, jid=jid)
            self._sessions[jingle.sid] = jingle
            video = JingleVideo(jingle)
            jingle.add_content("video", video)
            jingle.start_session()
        return jingle.sid

    def start_audio_video(self, jid: str) -> str:
        video_session = self.get_jingle_session(jid, media="video")
        if video_session is not None:
            return video_session.sid

        audio_session = self.get_jingle_session(jid, media="audio")
        video_session = self.get_jingle_session(jid, media="video")
        if audio_session and video_session:
            return audio_session.sid
        if audio_session:
            video = JingleVideo(audio_session)
            audio_session.add_content("video", video)
            return audio_session.sid
        if video_session:
            audio = JingleAudio(video_session)
            video_session.add_content("audio", audio)
            return video_session.sid

        jingle_session = JingleSession(self._con, weinitiate=True, jid=jid)
        self._sessions[jingle_session.sid] = jingle_session
        audio = JingleAudio(jingle_session)
        video = JingleVideo(jingle_session)
        jingle_session.add_content("audio", audio)
        jingle_session.add_content("video", video)
        jingle_session.start_session()
        return jingle_session.sid

    def start_file_transfer(
        self, jid: str, file_props: FileProp, request: bool = False
    ) -> str | None:
        logger.info("start file transfer with file: %s", file_props)
        contact = self._con.get_module("Contacts").get_contact(jid)
        jingle = JingleSession(self._con, weinitiate=True, jid=jid, werequest=request)
        # this is a file transfer
        jingle.session_type_ft = True
        self._sessions[jingle.sid] = jingle
        file_props.sid = jingle.sid

        if contact.supports(Namespace.JINGLE_BYTESTREAM):
            transport = JingleTransportSocks5()
        elif contact.supports(Namespace.JINGLE_IBB):
            transport = JingleTransportIBB()
        else:
            logger.error("No suitable transport method available for %s", contact.jid)
            return None

        senders = "initiator"
        if request:
            senders = "responder"
        transfer = JingleFileTransfer(
            jingle, transport=transport, file_props=file_props, senders=senders
        )
        file_props.transport_sid = transport.sid
        file_props.algo = self.__hash_support(contact)
        jingle.add_content("file" + get_random_string(), transfer)
        jingle.start_session()
        return transfer.transport.sid

    @staticmethod
    def __hash_support(contact: types.ResourceContact) -> str | None:
        if contact.supports(Namespace.HASHES_2):
            if contact.supports(Namespace.HASHES_BLAKE2B_512):
                return "blake2b-512"
            if contact.supports(Namespace.HASHES_BLAKE2B_256):
                return "blake2b-256"
            if contact.supports(Namespace.HASHES_SHA3_512):
                return "sha3-512"
            if contact.supports(Namespace.HASHES_SHA3_256):
                return "sha3-256"
            if contact.supports(Namespace.HASHES_SHA512):
                return "sha-512"
            if contact.supports(Namespace.HASHES_SHA256):
                return "sha-256"
        return None

    def get_jingle_sessions(
        self, jid: str | None, sid: str | None = None, media: str | None = None
    ) -> list[JingleSession]:
        if sid:
            return [se for se in self._sessions.values() if se.sid == sid]

        sessions = [se for se in self._sessions.values() if se.peerjid == jid]
        if media:
            if media not in ("audio", "video", "file"):
                return []
            return [se for se in sessions if se.get_content(media)]
        return sessions

    def set_file_info(self, file_: dict[str, Any]) -> None:
        # Saves information about the files we have transferred
        # in case they need to be requested again.
        self.files.append(file_)

    def get_file_info(
        self,
        peerjid: str,
        hash_: str | None = None,
        name: str | None = None,
        _account: str | None = None,
    ) -> dict[str, Any] | None:
        if hash_:
            for file in self.files:  # DEBUG
                # if f['hash'] == '1294809248109223':
                if file["hash"] == hash_ and file["peerjid"] == peerjid:
                    return file
        elif name:
            for file in self.files:
                if file["name"] == name and file["peerjid"] == peerjid:
                    return file
        return None

    def get_jingle_session(
        self, jid: str, sid: str | None = None, media: str | None = None
    ) -> JingleSession | None:
        if sid is not None:
            if sid in self._sessions:
                return self._sessions[sid]
            return None
        if media is not None:
            if media not in ("audio", "video", "file"):
                return None
            for session in self._sessions.values():
                if session.peerjid == jid and session.get_content(media):
                    return session
        return None



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import cast
from typing import Literal

from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from nbxmpp import JID
from nbxmpp import Namespace

from gajim.common import app
from gajim.common import events
from gajim.common import ged
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.modules.contacts import GroupchatContact

from gajim.gtk import structs
from gajim.gtk.alert import AlertDialog
from gajim.gtk.alert import CancelDialogResponse
from gajim.gtk.alert import DialogCheckButton
from gajim.gtk.alert import DialogResponse
from gajim.gtk.chat_filter import ChatFilter
from gajim.gtk.chat_list import ChatList
from gajim.gtk.chat_list_row import ChatListRow


class ChatListStack(Gtk.Stack, EventHelper):
    __gtype_name__ = "ChatListStack"
    __gsignals__ = {
        "unread-count-changed": (GObject.SignalFlags.RUN_LAST, None, (str, int)),
        "chat-selected": (GObject.SignalFlags.RUN_LAST, None, (str, str, object)),
        "chat-unselected": (GObject.SignalFlags.RUN_LAST, None, ()),
        "chat-removed": (GObject.SignalFlags.RUN_LAST, None, (str, object, str, bool)),
    }

    def __init__(self) -> None:
        Gtk.Stack.__init__(self)
        EventHelper.__init__(self)
        self.set_hexpand(True)
        self.set_vexpand(True)
        self.set_vhomogeneous(False)

        self._chat_lists: dict[str, ChatList] = {}

        self._last_visible_child_name: str = "default"

        self.add_named(Gtk.Box(), "default")

        self.connect("notify::visible-child-name", self._on_visible_child_name)

        self._add_actions()

        self.register_events(
            [
                ("message-received", ged.GUI2, self._on_event),
                ("message-corrected", ged.GUI2, self._on_event),
                ("message-moderated", ged.GUI2, self._on_event),
                ("message-retracted", ged.GUI2, self._on_event),
                ("presence-received", ged.GUI2, self._on_event),
                ("message-sent", ged.GUI2, self._on_event),
                ("message-deleted", ged.GUI2, self._on_event),
                ("file-request-received", ged.GUI2, self._on_event),
                ("jingle-request-received", ged.GUI2, self._on_event),
            ]
        )

    def set_search_entry(self, search_entry: Gtk.SearchEntry) -> None:
        search_entry.connect("search-changed", self._on_search_changed)

    def set_chat_filter(self, chat_filter: ChatFilter) -> None:
        chat_filter.connect("filter-changed", self._on_filter_changed)

    def _add_actions(self) -> None:
        actions = [
            ("toggle-chat-pinned", "a{sv}", self._toggle_chat_pinned),
            ("move-chat-to-workspace", "a{sv}", self._move_chat_to_workspace),
            ("mark-as-read", "a{sv}", self._mark_as_read),
        ]

        for action in actions:
            action_name, variant, func = action
            variant = GLib.VariantType.new(variant)
            act = Gio.SimpleAction.new(action_name, variant)
            act.connect("activate", func)
            app.window.add_action(act)

    def _on_visible_child_name(self, _stack: Gtk.Stack, _param: str) -> None:
        new_visible_child_name = self.get_visible_child_name()
        if self._last_visible_child_name == new_visible_child_name:
            return

        if self._last_visible_child_name != "default":
            chat_list = cast(
                ChatList | None, self.get_child_by_name(self._last_visible_child_name)
            )
            if chat_list is not None:
                chat_list.set_filter_text("")

        self._last_visible_child_name = new_visible_child_name or "default"

    def get_chatlist(self, workspace_id: str) -> ChatList:
        return self._chat_lists[workspace_id]

    def get_selected_chat(self) -> ChatListRow | None:
        chat_list = self.get_current_chat_list()
        if chat_list is None:
            return None
        return chat_list.get_selected_chat()

    def get_current_chat_list(self) -> ChatList | None:
        workspace_id = self.get_visible_child_name()
        if workspace_id == "default" or workspace_id is None:
            return None

        return self._chat_lists[workspace_id]

    def is_chat_selected(self, account: str, jid: JID) -> bool:
        chat = self.get_selected_chat()
        if chat is None:
            return False
        return chat.account == account and chat.jid == jid

    def _on_filter_changed(self, chat_filter: ChatFilter) -> None:
        filters = chat_filter.get_filters()

        chat_list = cast(ChatList, self.get_visible_child())
        chat_list.set_filter(filters)

    def _on_search_changed(self, search_entry: Gtk.SearchEntry) -> None:
        chat_list = cast(ChatList, self.get_visible_child())
        chat_list.set_filter_text(search_entry.get_text())

    def add_chat_list(self, workspace_id: str) -> ChatList:
        chat_list = ChatList(workspace_id)
        chat_list.connect("row-selected", self._on_row_selected)
        chat_list.connect("chat-order-changed", self._on_chat_order_changed)

        self._chat_lists[workspace_id] = chat_list
        self.add_named(chat_list, workspace_id)
        return chat_list

    def remove_chat_list(self, workspace_id: str) -> None:
        chat_list = self._chat_lists[workspace_id]
        self.remove(chat_list)
        for open_chat in chat_list.get_open_chats():
            self.remove_chat(workspace_id, open_chat["account"], open_chat["jid"])

        self._chat_lists.pop(workspace_id)

    def _on_row_selected(self, _chat_list: ChatList, row: ChatListRow | None) -> None:
        if row is None:
            self.emit("chat-unselected")
            return

        self.emit("chat-selected", row.workspace_id, row.account, row.jid)

    def _on_chat_order_changed(self, chat_list: ChatList) -> None:
        self.store_open_chats(chat_list.workspace_id)

    def show_chat_list(self, workspace_id: str) -> None:
        cur_workspace_id = self.get_visible_child_name()
        if cur_workspace_id != "default" and cur_workspace_id is not None:
            self._chat_lists[cur_workspace_id].unselect_all()

        self.set_visible_child_name(workspace_id)

    def add_chat(
        self,
        workspace_id: str,
        account: str,
        jid: JID,
        type_: Literal["chat", "groupchat", "pm"],
        pinned: bool,
        position: int,
    ) -> None:
        chat_list = self._chat_lists.get(workspace_id)
        if chat_list is None:
            chat_list = self.add_chat_list(workspace_id)
        chat_list.add_chat(account, jid, type_, pinned, position)

    def select_chat(self, account: str, jid: JID) -> None:
        chat_list = self.find_chat(account, jid)
        if chat_list is None:
            return

        self.show_chat_list(chat_list.workspace_id)
        chat_list.select_chat(account, jid)

    def store_open_chats(self, workspace_id: str) -> None:
        chat_list = self._chat_lists[workspace_id]
        open_chats = chat_list.get_open_chats()
        app.settings.set_workspace_setting(workspace_id, "chats", open_chats)

    @structs.actionmethod
    def _toggle_chat_pinned(
        self, _action: Gio.SimpleAction, params: structs.ChatListEntryParam
    ) -> None:
        chat_list = self._chat_lists[params.workspace_id]
        chat_list.toggle_chat_pinned(params.account, params.jid)
        self.store_open_chats(params.workspace_id)

    @structs.actionmethod
    def _move_chat_to_workspace(
        self, _action: Gio.SimpleAction, params: structs.ChatListEntryParam
    ) -> None:
        workspace_id = params.workspace_id
        if not workspace_id or workspace_id == "add-workspace":
            workspace_id = app.window.add_workspace(switch=False)

        source_chatlist = self.get_chatlist(params.source_workspace_id)
        type_ = source_chatlist.get_chat_type(params.account, params.jid)
        if type_ is None:
            return

        source_chatlist.remove_chat(params.account, params.jid)

        new_chatlist = self.get_chatlist(workspace_id)
        new_chatlist.add_chat(params.account, params.jid, type_, False, -1)

        self.store_open_chats(source_chatlist.workspace_id)
        self.store_open_chats(workspace_id)

    @structs.actionmethod
    def _mark_as_read(
        self, _action: Gio.SimpleAction, params: structs.AccountJidParam
    ) -> None:
        self.mark_as_read(params.account, params.jid)

    def remove_chat(self, workspace_id: str, account: str, jid: JID) -> None:
        chat_list = self._chat_lists[workspace_id]
        type_ = chat_list.get_chat_type(account, jid)

        def _leave(response_id: str, not_ask_again: bool) -> None:
            if not_ask_again:
                app.settings.set("confirm_close_muc", False)
            _remove(unregister=response_id == "leave_perm")

        def _remove(unregister: bool = False) -> None:
            chat_list.remove_chat(account, jid, emit_unread=False)
            self.store_open_chats(workspace_id)
            self.emit("chat-removed", account, jid, type_, unregister)

        if type_ != "groupchat" or not app.settings.get("confirm_close_muc"):
            _remove()
            return

        client = app.get_client(account)
        contact = client.get_module("Contacts").get_contact(jid)
        assert isinstance(contact, GroupchatContact)

        if contact.is_not_joined and client.is_available():
            # For example a chat opened from the search bar
            _remove()
            return

        muc_disco = contact.get_disco()
        us = contact.get_self()
        if muc_disco is None or us is None:
            _remove()
            return

        responses = [
            CancelDialogResponse(),
            DialogResponse("leave", _("_Leave")),
        ]

        affiliation = us.affiliation
        text = _('By closing this chat, you will leave "%s".') % contact.name
        if Namespace.REGISTER in muc_disco.features and not affiliation.is_none:
            responses.append(
                DialogResponse(
                    "leave_perm", _("Leave _Permanently"), appearance="destructive"
                )
            )

            text += "\n"
            text += _(
                "You may need an invite to join this group chat again, "
                "if you choose to leave it permanently."
            )
            if affiliation.is_admin:
                text += "\n"
                text += _("Additionally, you will lose your administrator affiliation.")
            elif affiliation.is_owner:
                text += "\n"
                text += _("Additionally, you will lose your owner affiliation.")

        AlertDialog(
            _("Leave Group Chat?"),
            text,
            responses=responses,
            emit_responses=["leave", "leave_perm"],
            extra_widget=DialogCheckButton(_("_Do not ask me again")),
            callback=_leave,
        )

    def remove_chats_for_account(self, account: str) -> None:
        for workspace_id, chat_list in self._chat_lists.items():
            chat_list.remove_chats_for_account(account)
            self.store_open_chats(workspace_id)

    def find_chat(self, account: str, jid: JID) -> ChatList | None:
        for chat_list in self._chat_lists.values():
            if chat_list.contains_chat(account, jid):
                return chat_list
        return None

    def contains_chat(
        self, account: str, jid: JID, workspace_id: str | None = None
    ) -> bool:
        if workspace_id is None:
            return any(
                chat_list.contains_chat(account, jid)
                for chat_list in self._chat_lists.values()
            )

        chat_list = self._chat_lists[workspace_id]
        return chat_list.contains_chat(account, jid)

    def get_total_unread_count(self) -> int:
        count = 0
        for chat_list in self._chat_lists.values():
            count += chat_list.get_unread_count()
        return count

    def get_chat_unread_count(
        self, account: str, jid: JID, include_silent: bool = False
    ) -> int | None:
        for chat_list in self._chat_lists.values():
            count = chat_list.get_chat_unread_count(account, jid, include_silent)
            if count is not None:
                return count
        return None

    def set_chat_unread_count(self, account: str, jid: JID, count: int) -> None:
        for chat_list in self._chat_lists.values():
            chat_list.set_chat_unread_count(account, jid, count)

    def mark_as_read(self, account: str, jid: JID) -> None:
        for chat_list in self._chat_lists.values():
            chat_list.mark_as_read(account, jid)

    def _on_event(self, event: events.ChatListEventT) -> None:
        jid = JID.from_string(event.jid)

        chat_list = self.find_chat(event.account, jid)
        if chat_list is None:
            return
        chat_list.process_event(event)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

import logging

from gi.repository import Gtk
from gi.repository import GtkSource
from nbxmpp.errors import StanzaError
from nbxmpp.modules import dataforms
from nbxmpp.simplexml import Node
from nbxmpp.task import Task

from gajim.common import app
from gajim.common import ged
from gajim.common import types
from gajim.common.const import SimpleClientState
from gajim.common.ged import EventHelper
from gajim.common.helpers import to_user_string
from gajim.common.i18n import _

from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.builder import get_builder
from gajim.gtk.dataform import DataFormWidget
from gajim.gtk.util.styling import get_source_view_style_scheme
from gajim.gtk.window import GajimAppWindow

log = logging.getLogger("gajim.gtk.pep_config")


class PEPConfig(GajimAppWindow, EventHelper):
    def __init__(self, account: str) -> None:
        GajimAppWindow.__init__(
            self,
            name="PEPConfig",
            title=_("PEP Service Configuration (%s)") % account,
            default_width=700,
            default_height=800,
            add_window_padding=True,
            header_bar=True,
        )

        EventHelper.__init__(self)

        self._ui = get_builder("pep_config.ui")
        self.set_child(self._ui.stack)

        self.account = account
        self._client = app.get_client(account)
        self._client.connect_signal("state-changed", self._on_client_state_changed)

        self._result_node: Node | None = None
        self._dataform_widget: DataFormWidget | None = None

        source_manager = GtkSource.LanguageManager.get_default()
        lang = source_manager.get_language("xml")
        self._ui.items_view.get_buffer().set_language(lang)

        style_scheme = get_source_view_style_scheme()
        if style_scheme is not None:
            self._ui.items_view.get_buffer().set_style_scheme(style_scheme)

        self._init_services()

        selection = self._ui.services_treeview.get_selection()
        self._connect(selection, "changed", self._on_services_selection_changed)

        self._connect(self._ui.reload_button, "clicked", self._reload_pep_entries)
        self._connect(
            self._ui.show_content_button, "clicked", self._on_show_content_clicked
        )
        self._connect(self._ui.delete_button, "clicked", self._on_delete_button_clicked)
        self._connect(
            self._ui.configure_button, "clicked", self._on_configure_button_clicked
        )
        self._connect(self._ui.items_back_button, "clicked", self._on_back_clicked)
        self._connect(self._ui.config_back_button, "clicked", self._on_back_clicked)
        self._connect(self._ui.save_button, "clicked", self._on_save_config_clicked)

        self.register_events([("style-changed", ged.GUI1, self._on_style_changed)])

    def _cleanup(self) -> None:
        self.unregister_events()

    def _on_client_state_changed(
        self, _client: types.Client, _signal_name: str, state: SimpleClientState
    ):
        self._ui.reload_button.set_sensitive(state.is_connected)
        if not state.is_connected:
            self._ui.configure_button.set_sensitive(False)
            self._ui.show_content_button.set_sensitive(False)
            self._ui.delete_button.set_sensitive(False)

    def _on_style_changed(self, *args: Any) -> None:
        style_scheme = get_source_view_style_scheme()
        if style_scheme is not None:
            self._ui.items_view.get_buffer().set_style_scheme(style_scheme)

    def _on_services_selection_changed(self, _selection: Gtk.TreeSelection) -> None:
        if not self._client.is_available():
            return

        self._ui.configure_button.set_sensitive(True)
        self._ui.show_content_button.set_sensitive(True)
        self._ui.delete_button.set_sensitive(True)

    def _init_services(self):
        # service, access_model, group
        self.treestore = Gtk.ListStore(str)
        self.treestore.set_sort_column_id(0, Gtk.SortType.ASCENDING)
        self._ui.services_treeview.set_model(self.treestore)

        col = Gtk.TreeViewColumn(title=_("Service"))
        col.set_sort_column_id(0)
        self._ui.services_treeview.append_column(col)

        cellrenderer_text = Gtk.CellRendererText()
        col.pack_start(cellrenderer_text, True)
        col.add_attribute(cellrenderer_text, "text", 0)

        self._reload_pep_entries()

    def _reload_pep_entries(self, *args: Any) -> None:
        self.treestore.clear()

        self._ui.configure_button.set_sensitive(False)
        self._ui.show_content_button.set_sensitive(False)
        self._ui.delete_button.set_sensitive(False)

        jid = self._client.get_own_jid().bare
        self._client.get_module("Discovery").disco_items(
            jid, callback=self._items_received
        )

    def _items_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except StanzaError as error:
            InformationAlertDialog(_("Error"), to_user_string(error))
            return

        jid = result.jid.bare
        for item in result.items:
            if item.jid == jid and item.node is not None:
                self.treestore.append([item.node])

    def _on_delete_button_clicked(self, _widget: Gtk.Button) -> None:
        selection = self._ui.services_treeview.get_selection()
        if not selection:
            return

        model, iter_ = selection.get_selected()
        assert isinstance(model, Gtk.ListStore)
        assert iter_
        node = model[iter_][0]

        def _on_response() -> None:
            self._client.get_module("PubSub").delete(
                node, callback=self._on_node_delete, user_data=node
            )

        ConfirmationAlertDialog(
            _("Delete Node?"),
            _("Do you want to delete this PEP node?"),
            confirm_label=_("_Delete"),
            appearance="destructive",
            callback=_on_response,
            parent=self,
        )

    def _on_node_delete(self, task: Task) -> None:
        node = task.get_user_data()

        try:
            task.finish()
        except StanzaError as error:
            InformationAlertDialog(
                _("PEP Node Not Removed"),
                _("PEP node %(node)s was not removed:\n%(message)s")
                % {"node": node, "message": error},
            )
            return

        model = self._ui.services_treeview.get_model()
        assert isinstance(model, Gtk.ListStore)
        iter_ = model.get_iter_first()
        while iter_:
            if model[iter_][0] == node:
                model.remove(iter_)
                break
            iter_ = model.iter_next(iter_)

    def _on_configure_button_clicked(self, _button: Gtk.Button) -> None:
        selection = self._ui.services_treeview.get_selection()
        if not selection:
            return
        model, iter_ = selection.get_selected()
        assert isinstance(model, Gtk.ListStore)
        assert iter_
        node = model[iter_][0]

        self._client.get_module("PubSub").get_node_configuration(
            node, callback=self._on_pep_config_received
        )

    def _on_pep_config_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except Exception:
            log.exception("Failed to retrieve config")
            return

        if self._dataform_widget is not None:
            self._ui.form_box.remove(self._dataform_widget)

        self._result_node = result.node

        form = dataforms.extend_form(node=result.form)
        self._dataform_widget = DataFormWidget(form)
        self._dataform_widget.set_propagate_natural_height(True)
        self._dataform_widget.set_visible(True)
        self._ui.form_box.append(self._dataform_widget)
        self._ui.form_label.set_text(result.node)

        self._ui.stack.set_visible_child_name("config")

    def _on_save_config_clicked(self, _button: Gtk.Button) -> None:
        assert self._dataform_widget is not None
        form = self._dataform_widget.get_submit_form()
        self._client.get_module("PubSub").set_node_configuration(
            self._result_node, form
        )

        self._ui.stack.set_visible_child_name("overview")

    def _on_back_clicked(self, _button: Gtk.Button) -> None:
        self._ui.stack.set_visible_child_name("overview")

    def _on_show_content_clicked(self, _button: Gtk.Button) -> None:
        selection = self._ui.services_treeview.get_selection()
        if not selection:
            return
        model, iter_ = selection.get_selected()
        assert isinstance(model, Gtk.ListStore)
        assert iter_
        node = model[iter_][0]

        self._client.get_module("PubSub").request_items(
            node, callback=self._on_pep_items_received, user_data=node
        )

    def _on_pep_items_received(self, task: Task) -> None:
        try:
            result = task.finish()
        except Exception:
            log.exception("Failed to retrieve items")
            return

        parent_node = task.get_user_data()
        self._ui.items_label.set_text(str(parent_node))

        buf = self._ui.items_view.get_buffer()
        start, end = buf.get_bounds()
        buf.delete(start, end)

        for node in result:
            # pylint: disable=unnecessary-dunder-call
            node = node.__str__(fancy=True)

            buf = self._ui.items_view.get_buffer()
            buf.insert_at_cursor(str(node))

        self._ui.stack.set_visible_child_name("items")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging

from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import Gtk
from nbxmpp.protocol import JID

from gajim.common import app
from gajim.common import ged
from gajim.common.const import AvatarSize
from gajim.common.const import CallType
from gajim.common.const import JingleState
from gajim.common.events import CallUpdated
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.jingle_rtp import JingleVideo
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import ResourceContact

from gajim.gtk.builder import get_builder
from gajim.gtk.gstreamer import create_video_elements
from gajim.gtk.window import GajimAppWindow

log = logging.getLogger("gajim.gtk.call_window")


class CallWindow(GajimAppWindow, EventHelper):
    def __init__(self, account: str, resource_jid: JID) -> None:
        GajimAppWindow.__init__(
            self,
            name="CallWindow",
            default_width=700,
            default_height=600,
            add_window_padding=True,
            header_bar=True,
        )

        EventHelper.__init__(self)

        self._account = account
        self._client = app.get_client(account)

        contacts_module = self._client.get_module("Contacts")
        self._contact = contacts_module.get_contact(resource_jid.bare)
        assert isinstance(self._contact, BareContact)
        self._resource_contact = contacts_module.get_contact(resource_jid)
        assert isinstance(self._resource_contact, ResourceContact)
        self.set_title(_("Call with %s") % self._contact.name)

        self._video_widget_other = None
        self._video_widget_self = None

        self._ui = get_builder("call_window.ui")
        self.set_child(self._ui.av_box)

        buttons = [
            self._ui.button_0,
            self._ui.button_1,
            self._ui.button_2,
            self._ui.button_3,
            self._ui.button_3,
            self._ui.button_4,
            self._ui.button_5,
            self._ui.button_6,
            self._ui.button_7,
            self._ui.button_8,
            self._ui.button_9,
            self._ui.button_star,
            self._ui.button_pound,
        ]
        for button in buttons:
            gesture_primary_click = Gtk.GestureClick(
                button=Gdk.BUTTON_PRIMARY,
                propagation_phase=Gtk.PropagationPhase.CAPTURE,
            )
            self._connect(gesture_primary_click, "pressed", self._on_button_press)
            self._connect(gesture_primary_click, "released", self._on_button_release)
            button.add_controller(gesture_primary_click)

        self._connect(
            self._ui.answer_video_button, "clicked", self._on_upgrade_to_video_clicked
        )
        self._connect(self._ui.av_cam_button, "clicked", self._on_enable_video_clicked)
        self._connect(self._ui.end_call_button, "clicked", self._on_end_call_clicked)
        self._connect(self._ui.mic_hscale, "value-changed", self._on_mic_volume_changed)
        self._connect(
            self._ui.sound_hscale, "value-changed", self._on_output_volume_changed
        )

        self._ui.avatar_image.set_pixel_size(AvatarSize.CALL_BIG)
        self._ui.avatar_image.set_from_paintable(
            self._contact.get_avatar(
                AvatarSize.CALL_BIG, self.get_scale_factor(), add_show=False
            )
        )

        self.register_events(
            [
                ("call-updated", ged.GUI2, self._on_call_updated),
            ]
        )

    def _cleanup(self) -> None:
        assert isinstance(self._resource_contact, ResourceContact)
        app.call_manager.stop_call(self._account, self._resource_contact)
        self.unregister_events()

    def _close_with_timeout(self, timeout: int = 3) -> None:
        self._ui.av_box.set_sensitive(False)
        GLib.timeout_add_seconds(timeout, self.close)

    def _on_upgrade_to_video_clicked(self, button: Gtk.Button) -> None:
        app.call_manager.upgrade_to_video_call()
        button.set_visible(False)

    def _on_end_call_clicked(self, button: Gtk.Button) -> None:
        button.set_sensitive(False)
        assert isinstance(self._resource_contact, ResourceContact)
        app.call_manager.stop_call(self._account, self._resource_contact)
        self._close_with_timeout(timeout=1)

    def _on_enable_video_clicked(self, _button: Gtk.Button) -> None:
        app.call_manager.start_call(self._account, self._contact.jid, CallType.VIDEO)

    def _on_button_press(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        button = gesture_click.get_widget()
        if button is None:
            return

        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)
        key = button.get_name()
        app.call_manager.start_dtmf(self._account, self._resource_contact.jid, key)

    def _on_button_release(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        _x: float,
        _y: float,
    ) -> None:
        app.call_manager.stop_dtmf(self._account, self._resource_contact.jid)

    def _on_mic_volume_changed(self, _button: Gtk.VolumeButton, value: float) -> None:
        app.call_manager.mic_volume_changed(
            self._account, self._resource_contact.jid, int(value)
        )

    def _on_output_volume_changed(
        self, _button: Gtk.VolumeButton, value: float
    ) -> None:
        app.call_manager.output_volume_changed(
            self._account, self._resource_contact.jid, int(value)
        )

    def _on_call_updated(self, event: CallUpdated) -> None:
        if event.jingle_type == "audio":
            self._update_audio(event)
        else:
            self._update_video(event)

    def _update_audio(self, event: CallUpdated) -> None:
        assert isinstance(self._contact, BareContact)
        if self._contact.supports_video:
            self._ui.av_cam_button.set_sensitive(
                event.video_state not in (JingleState.CONNECTING, JingleState.CONNECTED)
            )

        if event.audio_state == JingleState.NULL:
            self._ui.audio_buttons_box.set_sensitive(False)
            self._ui.jingle_audio_state.set_visible(False)
            self._ui.jingle_audio_state.set_visible(False)
            self._ui.jingle_connection_state.set_text("")
            self._ui.jingle_connection_spinner.set_visible(False)

            if event.video_state == JingleState.NULL:
                self._ui.jingle_connection_state.set_text(_("Call ended"))
                self._close_with_timeout()
        else:
            self._ui.jingle_connection_spinner.set_visible(True)

        if event.audio_state == JingleState.CONNECTING:
            self._ui.jingle_connection_state.set_text(_("Calling…"))
            self._ui.av_cam_button.set_sensitive(False)

        elif event.audio_state == JingleState.CONNECTION_RECEIVED:
            self._ui.jingle_connection_state.set_text(_("Incoming Call"))

        elif event.audio_state == JingleState.CONNECTED:
            self._ui.jingle_audio_state.set_visible(True)
            self._ui.jingle_connection_state.set_text("")
            self._ui.jingle_connection_spinner.set_visible(False)
            if self._contact.supports_video:
                self._ui.av_cam_button.set_sensitive(True)

            input_vol = app.settings.get("audio_input_volume")
            output_vol = app.settings.get("audio_output_volume")
            self._ui.mic_hscale.set_value(max(min(input_vol, 100), 0))
            self._ui.sound_hscale.set_value(max(min(output_vol, 100), 0))
            self._ui.audio_buttons_box.set_sensitive(True)

        elif event.audio_state == JingleState.ERROR:
            self._ui.jingle_audio_state.set_visible(False)
            self._ui.jingle_connection_state.set_text(_("Connection Error"))
            self._ui.jingle_connection_spinner.set_visible(False)

        if not event.audio_sid:
            self._ui.audio_buttons_box.set_sensitive(False)

    def _update_video(self, event: CallUpdated) -> None:
        if event.video_state == JingleState.NULL:
            self._ui.avatar_image.set_visible(True)
            self._ui.video_box.set_visible(False)
            self._ui.video_box.set_visible(False)
            self._ui.outgoing_viewport.set_visible(False)
            self._ui.outgoing_viewport.set_visible(False)

            if self._video_widget_other:
                self._ui.incoming_viewport.set_child(None)
                self._video_widget_other = None

            if self._video_widget_self:
                self._ui.outgoing_viewport.set_child(None)
                self._video_widget_self = None

            if event.audio_state != JingleState.CONNECTED:
                self._ui.jingle_connection_state.set_text("")
            self._ui.jingle_connection_spinner.set_visible(False)
            self._ui.av_cam_button.set_sensitive(True)
            self._ui.av_cam_button.set_tooltip_text(_("Turn Camera on"))
            self._ui.av_cam_image.set_from_icon_name("lucide-camera-symbolic")
            if event.audio_state == JingleState.NULL:
                self._ui.jingle_connection_state.set_text(_("Call ended"))
                self._close_with_timeout()
        else:
            self._ui.jingle_connection_spinner.set_visible(True)

        if event.video_state == JingleState.CONNECTING:
            self._ui.jingle_connection_state.set_text(_("Calling (Video)…"))
            self._ui.av_cam_button.set_sensitive(False)
            self._ui.av_cam_button.set_tooltip_text(_("Turn Camera off"))
            self._ui.av_cam_image.set_from_icon_name("lucide-camera-off-symbolic")

        elif event.video_state == JingleState.CONNECTION_RECEIVED:
            self._ui.jingle_connection_state.set_text(_("Incoming Call (Video)"))
            self._ui.answer_video_button.set_visible(True)
            self._ui.av_cam_button.set_sensitive(False)
            self._ui.av_cam_button.set_tooltip_text(_("Turn Camera off"))
            self._ui.av_cam_image.set_from_icon_name("lucide-camera-off-symbolic")

        elif event.video_state == JingleState.CONNECTED:
            self._ui.avatar_image.set_visible(False)
            self._ui.answer_video_button.set_visible(False)
            if app.settings.get("video_see_self"):
                self._ui.outgoing_viewport.set_visible(True)
            else:
                self._ui.outgoing_viewport.set_visible(False)
                self._ui.outgoing_viewport.set_visible(False)

            other_video_elements = create_video_elements()
            self_video_elements = create_video_elements()
            if other_video_elements is None or self_video_elements is None:
                log.warning("Could not create GStreamer widgets")
                return

            sink_other, paintable_other, _name = other_video_elements
            sink_self, paintable_self, _name = self_video_elements
            self._video_widget_other = Gtk.Picture(paintable=paintable_other)
            self._video_widget_self = Gtk.Picture(paintable=paintable_self)
            self._ui.incoming_viewport.set_child(self._video_widget_other)
            self._ui.outgoing_viewport.set_child(self._video_widget_self)

            session = self._client.get_module("Jingle").get_jingle_session(
                str(self._resource_contact.jid), event.video_sid
            )
            assert session is not None
            content = session.get_content("video")
            assert isinstance(content, JingleVideo)
            content.do_setup(sink_self, sink_other)

            self._ui.jingle_connection_state.set_text("")
            self._ui.jingle_connection_spinner.set_visible(False)

            self._ui.av_cam_button.set_sensitive(True)
            self._ui.av_cam_button.set_tooltip_text(_("Turn Camera off"))
            self._ui.av_cam_image.set_from_icon_name("lucide-camera-off-symbolic")

        elif event.video_state == JingleState.ERROR:
            self._ui.jingle_connection_state.set_text(_("Connection Error"))
            self._ui.jingle_connection_spinner.set_visible(False)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gtk

from gajim.common import app

from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import check_finalize

log = logging.getLogger("gajim.gtk.window")


class GajimAppWindow(Adw.ApplicationWindow, SignalManager):
    __gtype_name__ = "GajimAppWindow"

    def __init__(
        self,
        *,
        name: str,
        title: str | None = None,
        default_width: int = 0,
        default_height: int = 0,
        transient_for: Gtk.Window | None = None,
        modal: bool = False,
        add_window_padding: bool = False,
        header_bar: bool = False,
    ) -> None:
        SignalManager.__init__(self)

        self._add_window_padding = add_window_padding

        self._header_bar = None
        if header_bar:
            self._header_bar = Adw.HeaderBar()

        window_size = app.settings.get_window_size(name)
        if window_size is not None:
            default_width, default_height = window_size

        Adw.ApplicationWindow.__init__(
            self,
            application=app.app,
            resizable=True,
            name=name,
            title=title,
            default_width=default_width,
            default_height=default_height,
            transient_for=transient_for,
            modal=modal,
        )

        log.debug("Load Window: %s", name)

        self.add_css_class("gajim-app-window")

        self.__default_controller = Gtk.EventControllerKey(
            propagation_phase=Gtk.PropagationPhase.CAPTURE
        )
        self.add_controller(self.__default_controller)

        self._connect_after(
            self.__default_controller, "key-pressed", self.__on_key_pressed
        )
        self._connect_after(self, "close-request", self.__on_close_request)

    def set_child(self, child: Gtk.Widget | None = None) -> None:
        if child is not None and self._add_window_padding:
            child.add_css_class("window-padding")

        if self._header_bar is not None:
            toolbar_view = Adw.ToolbarView(content=child)
            toolbar_view.add_top_bar(self._header_bar)
            self.set_content(toolbar_view)
        else:
            self.set_content(child)

    def get_default_controller(self) -> Gtk.EventController:
        return self.__default_controller

    def get_header_bar(self) -> Adw.HeaderBar | None:
        return self._header_bar

    def __on_key_pressed(
        self,
        _event_controller_key: Gtk.EventControllerKey,
        keyval: int,
        keycode: int,
        state: Gdk.ModifierType,
    ) -> bool:
        if keyval == Gdk.KEY_Escape:
            self.close()
            return Gdk.EVENT_STOP
        return Gdk.EVENT_PROPAGATE

    def __on_close_request(self, _widget: Adw.ApplicationWindow) -> bool:
        log.debug("Initiate Cleanup: %s", self.get_name())
        self._store_win_size()
        self._disconnect_all()
        self._cleanup()
        check_finalize(self)

        del self.__default_controller

        return Gdk.EVENT_PROPAGATE

    def _store_win_size(self) -> None:
        app.settings.set_window_size(
            self.get_name(),
            self.props.default_width,
            self.props.default_height,
        )

    def _cleanup(self) -> None:
        raise NotImplementedError



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any
from typing import cast
from typing import NamedTuple

import csv
import datetime
import enum
import logging
from pathlib import Path

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Pango
from nbxmpp.protocol import JID
from nbxmpp.structs import RosterItem

from gajim.common import app
from gajim.common import ged
from gajim.common.client import Client
from gajim.common.const import SimpleClientState
from gajim.common.events import RosterPush
from gajim.common.ged import EventHelper
from gajim.common.i18n import _
from gajim.common.util.decorators import event_filter

from gajim.gtk.alert import ConfirmationAlertDialog
from gajim.gtk.alert import DialogEntry
from gajim.gtk.alert import InformationAlertDialog
from gajim.gtk.menus import get_manage_roster_import_menu
from gajim.gtk.menus import get_manage_roster_menu
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.widgets import GajimPopover

log = logging.getLogger("gajim.gtk.manage_roster")


class CSVColumn(enum.IntEnum):
    JID = 0
    NAME = 1
    GROUPS = 2


class ImportedItem(NamedTuple):
    jid: JID
    name: str | None
    groups: set[str]


@Gtk.Template.from_string(string=get_ui_string("preference/manage_roster.ui"))
class ManageRoster(Gtk.Box, SignalManager, EventHelper):
    __gtype_name__ = "ManageRoster"

    _top_box: Gtk.Box = Gtk.Template.Child()
    _search_entry: Gtk.SearchEntry = Gtk.Template.Child()
    _import_button: Gtk.MenuButton = Gtk.Template.Child()
    _export_button: Gtk.Button = Gtk.Template.Child()
    _scrolled_box: Gtk.Box = Gtk.Template.Child()
    _scrolled: Gtk.ScrolledWindow = Gtk.Template.Child()
    _column_view: Gtk.ColumnView = Gtk.Template.Child()
    _jid_col: Gtk.ColumnViewColumn = Gtk.Template.Child()
    _name_col: Gtk.ColumnViewColumn = Gtk.Template.Child()
    _subscription_col: Gtk.ColumnViewColumn = Gtk.Template.Child()
    _ask_col: Gtk.ColumnViewColumn = Gtk.Template.Child()

    def __init__(self, account: str) -> None:
        Gtk.Box.__init__(self)
        EventHelper.__init__(self)
        SignalManager.__init__(self)

        self.account = account
        self._client = app.get_client(account)
        self._actions: list[tuple[str, Any, Any]] = []

        self._model = Gio.ListStore(item_type=RosterListItem)

        h_factory = Gtk.SignalListItemFactory()
        self._connect(h_factory, "setup", self._on_h_factory_setup)
        self._connect(h_factory, "bind", self._on_h_factory_bind)
        self._connect(h_factory, "unbind", self._on_factory_unbind)
        self._column_view.set_header_factory(h_factory)

        for _jid, item in self._client.get_module("Roster").iter():
            groups = item.groups or {""}
            for group in groups:
                self._model.append(RosterListItem(item, group))

        columns = [
            (self._jid_col, RosterViewItemLabel, "jid"),
            (self._name_col, RosterViewItemLabel, "name"),
            (self._subscription_col, RosterViewItemImage, "subscription"),
            (self._ask_col, RosterViewItemImage, "ask"),
        ]

        for col, widget, attr in columns:
            factory = col.get_factory()
            assert factory is not None
            self._connect(factory, "setup", self._on_factory_setup, widget)
            self._connect(factory, "bind", self._on_factory_bind, attr)
            self._connect(factory, "unbind", self._on_factory_unbind)

        expression = Gtk.PropertyExpression.new(
            this_type=RosterListItem,
            expression=None,
            property_name="group",
        )
        section_sorter = Gtk.StringSorter.new(expression=expression)

        expression = Gtk.PropertyExpression.new(
            this_type=RosterListItem,
            expression=None,
            property_name="jid",
        )
        sorter = Gtk.StringSorter.new(expression=expression)

        sort_model = Gtk.SortListModel(
            model=self._model, sorter=sorter, section_sorter=section_sorter
        )

        expression = Gtk.PropertyExpression.new(
            this_type=RosterListItem,
            expression=None,
            property_name="search_string",
        )

        self._string_filter = Gtk.StringFilter.new(expression)

        filter_model = Gtk.FilterListModel(model=sort_model, filter=self._string_filter)

        self._selection_model = Gtk.MultiSelection(model=filter_model)
        self._column_view.set_model(self._selection_model)

        self._popover_menu = GajimPopover(None)
        self._scrolled_box.append(self._popover_menu)

        gesture_secondary_click = Gtk.GestureClick(
            button=Gdk.BUTTON_SECONDARY, propagation_phase=Gtk.PropagationPhase.BUBBLE
        )
        self._connect(gesture_secondary_click, "pressed", self._popup_menu)
        self._scrolled_box.add_controller(gesture_secondary_click)

        accounts = app.get_enabled_accounts_with_labels()
        accounts = [(acc, label) for acc, label in accounts if acc != self.account]

        self._import_button.set_menu_model(
            get_manage_roster_import_menu(account, accounts)
        )
        self._export_button.set_action_name(f"win.{account}-export-to-csv")

        self._connect(self._search_entry, "search-changed", self._on_search_changed)

        self.register_event("roster-push", ged.GUI2, self._on_roster_push)
        self._client.connect_signal("state-changed", self._on_client_state_changed)

    def do_unroot(self) -> None:
        del self._selection_model
        del self._popover_menu
        del self._model
        del self._actions
        del self._string_filter
        self._client.disconnect_all_from_obj(self)
        self.unregister_events()
        self._disconnect_all()
        Gtk.Box.do_unroot(self)

    def do_root(self) -> None:
        Gtk.Box.do_root(self)
        self._add_actions(self.account)

    def _add_actions(self, account: str) -> None:
        self._actions = [
            (f"{account}-add-to-group", "s", self._on_add_to_group),
            (f"{account}-add-to-new-group", None, self._on_add_to_new_group),
            (f"{account}-move-to-group", "s", self._on_move_to_group),
            (f"{account}-move-to-new-group", None, self._on_move_to_new_group),
            (f"{account}-remove-from-group", None, self._remove_from_group),
            (f"{account}-remove-from-roster", None, self._on_remove_from_roster),
            (f"{account}-change-name", None, self._on_change_name),
            (f"{account}-import-from-account", "s", self._on_import_from_account),
            (f"{account}-import-from-file", None, self._on_import_from_file),
            (f"{account}-export-to-csv", None, self._on_export_to_csv),
        ]

        parent = self.get_ancestor(Adw.ApplicationWindow)
        assert parent is not None

        for action, variant_type, callback in self._actions:
            if variant_type is not None:
                variant_type = GLib.VariantType(variant_type)
            act = Gio.SimpleAction.new(action, variant_type)
            self._connect(act, "activate", callback)
            parent.add_action(act)

    @staticmethod
    def _on_factory_setup(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem, item_class: Any
    ) -> None:
        list_item.set_child(item_class())

    @staticmethod
    def _on_factory_bind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem, attr: str
    ) -> None:
        roster_item = cast(RosterListItem, list_item.get_item())
        roster_view_item = cast(
            RosterViewItemLabel | RosterViewItemImage, list_item.get_child()
        )
        roster_view_item.bind(roster_item, attr)

    @staticmethod
    def _on_factory_unbind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        roster_view_item = cast(
            RosterViewItemLabel | RosterViewItemImage, list_item.get_child()
        )
        roster_view_item.unbind()

    @staticmethod
    def _on_h_factory_setup(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        list_item.set_child(HeaderViewItem())

    @staticmethod
    def _on_h_factory_bind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        roster_item = cast(RosterListItem, list_item.get_item())
        roster_view_item = cast(HeaderViewItem, list_item.get_child())
        roster_view_item.bind(roster_item)

    @staticmethod
    def _on_h_factory_unbind(
        _factory: Gtk.SignalListItemFactory, list_item: Gtk.ListItem
    ) -> None:
        roster_view_item: HeaderViewItem = cast(HeaderViewItem, list_item.get_child())
        roster_view_item.unbind()

    @event_filter(["account"])
    def _on_roster_push(self, event: RosterPush) -> None:
        pos = 0
        while item := cast(RosterListItem, self._model.get_item(pos)):
            if item.jid == event.item.jid:
                self._model.remove(pos)
            else:
                pos += 1

        if event.item.subscription == "remove":
            return

        groups = event.item.groups or {""}
        for group in groups:
            self._model.append(RosterListItem(event.item, group))

    def _on_client_state_changed(
        self, _client: Client, _signal_name: str, state: SimpleClientState
    ) -> None:
        parent = self.get_ancestor(Adw.ApplicationWindow)
        assert parent is not None

        for actions in self._actions:
            action = parent.lookup_action(actions[0])
            assert isinstance(action, Gio.SimpleAction)
            action.set_enabled(state.is_connected)

    def _on_search_changed(self, search_entry: Gtk.SearchEntry) -> None:
        self._string_filter.set_search(search_entry.get_text())

    def _on_add_to_group(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:
        self._add_to_group(param.get_string())

    def _on_add_to_new_group(self, _action: Gio.SimpleAction, param: None) -> None:
        ConfirmationAlertDialog(
            _("Add Contact to New Group?"),
            _("Enter group name"),
            confirm_label=_("Add"),
            extra_widget=DialogEntry(text=_("New Group")),
            callback=self._add_to_group,
        )

    def _add_to_group(self, group: str) -> None:
        items = self._get_selected_items()

        for item in items:
            self._client.get_module("Roster").add_to_group(
                JID.from_string(item.jid), group
            )

    def _on_move_to_group(self, _action: Gio.SimpleAction, param: GLib.Variant) -> None:
        self._move_to_group(param.get_string())

    def _on_move_to_new_group(self, _action: Gio.SimpleAction, param: None) -> None:
        ConfirmationAlertDialog(
            _("Move Contact to New Group?"),
            _("Enter group name"),
            confirm_label=_("Move"),
            extra_widget=DialogEntry(text=_("New Group")),
            callback=self._move_to_group,
        )

    def _move_to_group(self, group: str) -> None:
        items = self._get_selected_items()

        for item in items:
            self._client.get_module("Roster").change_group(
                JID.from_string(item.jid), item.group, group
            )

    def _remove_from_group(self, _action: Gio.SimpleAction, param: None) -> None:
        items = self._get_selected_items()

        for item in items:
            self._client.get_module("Roster").remove_from_group(
                JID.from_string(item.jid), item.group
            )

    def _on_remove_from_roster(self, _action: Gio.SimpleAction, param: None) -> None:
        items = self._get_selected_items()

        def _on_response() -> None:
            if not self._client.is_available():
                return

            for item in items:
                self._client.get_module("Roster").delete_item(JID.from_string(item.jid))

        ConfirmationAlertDialog(
            _("Remove Contacts?"),
            _("Remove %s contacts from your contact list?") % len(items),
            confirm_label=_("_Remove"),
            appearance="destructive",
            callback=_on_response,
        )

    def _on_change_name(self, _action: Gio.SimpleAction, param: None) -> None:
        item = self._get_selected_items()[0]

        def _on_response(name: str) -> None:
            if not self._client.is_available():
                return
            self._client.get_module("Roster").change_name(
                JID.from_string(item.jid), name
            )

        ConfirmationAlertDialog(
            _("Rename Contact?"),
            _("Enter new contact name"),
            confirm_label=_("Rename"),
            extra_widget=DialogEntry(text=item.name),
            callback=_on_response,
        )

    def _on_import_from_account(
        self, _action: Gio.SimpleAction, param: GLib.Variant
    ) -> None:
        remote_client = app.get_client(param.get_string())
        remote_items = [
            item for _jid, item in remote_client.get_module("Roster").iter()
        ]

        if not remote_items:
            InformationAlertDialog(_("Import Error"), _("No contacts found to import"))
            return

        def _on_response() -> None:
            if not self._client.is_available():
                return

            for item in remote_items:
                self._client.get_module("Presence").subscribe(
                    item.jid, name=item.name, groups=item.groups, auto_auth=True
                )

        ConfirmationAlertDialog(
            _("Import Contacts?"),
            _("Found %s contacts to import") % len(remote_items),
            confirm_label=_("_Import"),
            appearance="suggested",
            callback=_on_response,
        )

    def _on_import_from_file(self, _action: Gio.SimpleAction, param: None) -> None:
        def _on_file_picked(dialog: Gtk.FileDialog, result: Gio.AsyncResult) -> None:
            try:
                g_file = dialog.open_finish(result)
            except GLib.Error as e:
                if e.code == 2:
                    # User dismissed dialog, do nothing
                    return

                log.exception(e)
                return

            path = g_file.get_path()
            assert path is not None
            self._import(Path(path))

        dialog = Gtk.FileDialog(
            default_filter=Gtk.FileFilter(name=_("CSV files"), suffixes=["csv"]),
            title=_("Import Contacts"),
            accept_label=_("Import"),
            modal=True,
        )

        parent = app.app.get_active_window()
        dialog.open(parent, None, _on_file_picked)

    def _on_export_to_csv(self, _action: Gio.SimpleAction, param: None) -> None:
        def _on_file_picked(dialog: Gtk.FileDialog, result: Gio.AsyncResult) -> None:
            try:
                g_file = dialog.save_finish(result)
            except GLib.Error as e:
                if e.code == 2:
                    # User dismissed dialog, do nothing
                    return

                log.exception(e)
                return

            path = g_file.get_path()
            assert path is not None
            self._export(Path(path))

        now = datetime.datetime.now()
        filename = f"{self.account}-{now.strftime('%Y-%m-%d')}.csv"

        dialog = Gtk.FileDialog(
            default_filter=Gtk.FileFilter(name=_("CSV files"), suffixes=["csv"]),
            title=_("Export CSV"),
            accept_label=_("Export"),
            initial_file=Gio.File.new_for_path(filename),
            modal=True,
        )

        parent = app.app.get_active_window()
        dialog.save(parent, None, _on_file_picked)

    def _export(self, path: Path) -> None:
        try:
            with path.open(mode="w", encoding="utf-8") as csvfile:
                writer = csv.writer(csvfile)
                for jid, item in self._client.get_module("Roster").iter():
                    writer.writerow([jid, item.name, ";".join(item.groups)])
        except PermissionError:
            InformationAlertDialog(
                _("Export Error"), _("No permission to export file to %s") % path.parent
            )

    def _import(self, path: Path) -> None:
        jids = {str(jid) for jid, _item in self._client.get_module("Roster").iter()}

        items: list[ImportedItem] = []
        with path.open(encoding="utf-8") as csvfile:
            reader = csv.reader(csvfile, strict=True)
            try:
                for row in reader:
                    if len(row) != 3:
                        continue

                    if row[CSVColumn.JID] in jids:
                        continue

                    item = self._validate_imported_item(row)
                    if item is not None:
                        items.append(item)
            except Exception:
                log.error("Could not read file %s to import roster contacts", path.name)
                InformationAlertDialog(
                    _("Import Error"),
                    _("Could not read content of %s. Only CSV files can be imported.")
                    % path.name,
                )
                return

        if not items:
            InformationAlertDialog(_("Import Error"), _("No contacts found to import"))
            return

        def _on_response() -> None:
            if not self._client.is_available():
                return

            for item in items:
                self._client.get_module("Presence").subscribe(
                    item.jid, name=item.name, groups=item.groups, auto_auth=True
                )

        ConfirmationAlertDialog(
            _("Import Contacts?"),
            _("Found %s contacts to import") % len(items),
            confirm_label=_("_Import"),
            appearance="suggested",
            callback=_on_response,
        )

    def _validate_imported_item(self, row: list[str]) -> ImportedItem | None:
        try:
            jid = JID.from_string(row[CSVColumn.JID])
        except Exception:
            log.warning("Invalid jid: %s", row[CSVColumn.JID])
            return None

        name = row[CSVColumn.NAME] or None

        groups: set[str] = set()
        if row[CSVColumn.GROUPS]:
            groups = set(row[CSVColumn.GROUPS].split(";"))

        return ImportedItem(jid=jid, name=name, groups=groups)

    def _get_selected_items(self) -> list[RosterListItem]:
        bitset = self._selection_model.get_selection()
        valid, iter_, value = Gtk.BitsetIter.init_first(bitset)

        items: list[RosterListItem] = []

        if not valid:
            return items

        items.append(cast(RosterListItem, self._selection_model.get_item(value)))

        while res := iter_.next():
            valid, value = res
            if not valid:
                break

            items.append(cast(RosterListItem, self._selection_model.get_item(value)))

        return items

    def _popup_menu(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        gesture_click.set_state(Gtk.EventSequenceState.CLAIMED)

        items = self._get_selected_items()
        if not items:
            return

        single_selection = len(items) == 1
        groups = self._client.get_module("Roster").get_groups()
        groups = sorted(groups)

        menu = get_manage_roster_menu(self.account, groups, single_selection)
        self._popover_menu.set_menu_model(menu)
        self._popover_menu.set_pointing_to_coord(x, y)
        self._popover_menu.popup()


class RosterListItem(GObject.Object):
    __gtype_name__ = "RosterListItem"

    jid = GObject.Property(type=str)
    name = GObject.Property(type=str)
    ask = GObject.Property(type=object)
    subscription = GObject.Property(type=object)
    group = GObject.Property(type=str)
    search_string = GObject.Property(type=str)

    def __init__(self, item: RosterItem, group: str) -> None:
        subscription = self._get_subscription_data(item.subscription)
        ask = self._get_ask_data(item.ask)

        super().__init__(
            jid=str(item.jid),
            name=item.name or "",
            ask=ask,
            subscription=subscription,
            group=group,
            search_string=f"{item.jid}|{item.name or ''}",
        )

    def _get_subscription_data(self, subscription: str | None) -> tuple[str, str, str]:
        match subscription:
            case "both":
                return (
                    "lucide-arrow-right-left-symbolic",
                    "success",
                    _("Status is shared"),
                )
            case "to":
                return (
                    "lucide-arrow-right-symbolic",
                    "warning",
                    _("You don't share your status with this contact"),
                )
            case "from":
                return (
                    "lucide-arrow-left-symbolic",
                    "accent",
                    _("This contact does not share their status with you"),
                )
            case "none":
                return ("lucide-x-symbolic", "error", _("Status is not shared"))
            case _:
                raise ValueError("Invalid value: %s" % subscription)

    def _get_ask_data(
        self, ask: str | None
    ) -> tuple[str | None, str | None, str | None]:
        match ask:
            case None:
                return (None, None, None)
            case "subscribe":
                return ("lucide-clock-symbolic", None, _("Request pending"))
            case _:
                raise ValueError("Invalid value: %s" % ask)

    def __repr__(self) -> str:
        return f"RosterListItem: {self.jid} {self.name}"


class RosterViewItemLabel(Gtk.Label):
    __gtype_name__ = "RosterViewItemLabel"

    def __init__(self) -> None:
        Gtk.Label.__init__(
            self,
            ellipsize=Pango.EllipsizeMode.END,
            width_chars=20,
            xalign=0,
        )

        self.__bindings: list[GObject.Binding] = []

    def bind(self, obj: RosterListItem, attr: str) -> None:
        bind_spec = [
            (attr, self, "label"),
        ]

        for source_prop, widget, target_prop in bind_spec:
            bind = obj.bind_property(
                source_prop, widget, target_prop, GObject.BindingFlags.SYNC_CREATE
            )
            self.__bindings.append(bind)

    def unbind(self) -> None:
        for bind in self.__bindings:
            bind.unbind()
        self.__bindings.clear()

    def do_unroot(self) -> None:
        Gtk.Label.do_unroot(self)
        app.check_finalize(self)


class RosterViewItemImage(Gtk.Image):
    __gtype_name__ = "RosterViewItemImage"

    def __init__(self) -> None:
        Gtk.Image.__init__(
            self,
            pixel_size=16,
        )

        self._icon_data = ("", "", "")
        self.__bindings: list[GObject.Binding] = []

    @GObject.Property(type=object)
    def icon_data(self) -> tuple[str | None, str | None, str | None]:  # type: ignore
        return self._icon_data

    @icon_data.setter
    def icon_data(self, value: tuple[str | None, str | None, str | None]) -> None:
        self._icon_data = value
        icon_name, css_class, tooltip = value

        self.set_from_icon_name(icon_name)
        self.set_tooltip_text(tooltip)

        self.remove_css_class("success")
        self.remove_css_class("warning")
        self.remove_css_class("error")
        self.remove_css_class("accent")
        if css_class is not None:
            self.add_css_class(css_class)

    def bind(self, obj: RosterListItem, attr: str) -> None:
        bind_spec = [
            (attr, self, "icon-data"),
        ]

        for source_prop, widget, target_prop in bind_spec:
            bind = obj.bind_property(
                source_prop, widget, target_prop, GObject.BindingFlags.SYNC_CREATE
            )
            self.__bindings.append(bind)

    def unbind(self) -> None:
        for bind in self.__bindings:
            bind.unbind()
        self.__bindings.clear()

    def do_unroot(self) -> None:
        Gtk.Image.do_unroot(self)
        app.check_finalize(self)


class HeaderViewItem(Gtk.Label):
    __gtype_name__ = "HeaderViewItem"

    def __init__(self) -> None:
        Gtk.Label.__init__(
            self,
            ellipsize=Pango.EllipsizeMode.END,
            halign=Gtk.Align.START,
            xalign=0,
        )

        self._group_data = ""
        self.__bindings: list[GObject.Binding] = []

    @GObject.Property(type=str)
    def group_data(self) -> str:  # type: ignore
        return self._group_data

    @group_data.setter
    def group_data(self, value: str) -> None:
        self._group_data = value or _("No group assigned")
        self.set_label(self._group_data)

    def bind(self, obj: RosterListItem) -> None:
        bind_spec = [
            ("group", self, "group_data"),
        ]

        for source_prop, widget, target_prop in bind_spec:
            bind = obj.bind_property(
                source_prop, widget, target_prop, GObject.BindingFlags.SYNC_CREATE
            )
            self.__bindings.append(bind)

    def unbind(self) -> None:
        for bind in self.__bindings:
            bind.unbind()
        self.__bindings.clear()

    def do_unroot(self) -> None:
        Gtk.Label.do_unroot(self)
        app.check_finalize(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# Message Reactions (XEP-0444)

from __future__ import annotations

from nbxmpp.namespaces import Namespace
from nbxmpp.protocol import Message
from nbxmpp.protocol import NodeProcessed
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.events import ReactionUpdated
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.modules.message_util import get_chat_type_and_direction
from gajim.common.modules.message_util import get_message_timestamp
from gajim.common.modules.message_util import get_occupant_info
from gajim.common.storage.archive import models as mod
from gajim.common.structs import MessageType
from gajim.common.structs import OutgoingMessage
from gajim.common.util.text import convert_to_codepoints
from gajim.common.util.text import normalize_reactions


class Reactions(BaseModule):
    _nbxmpp_extends = "Reactions"

    def __init__(self, client: types.Client) -> None:
        BaseModule.__init__(self, client)

        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._process_reaction,
                ns=Namespace.REACTIONS,
                priority=47,
            )
        ]

    def _process_reaction(
        self,
        _client: types.NBXMPPClient,
        _stanza: Message,
        properties: MessageProperties,
    ) -> None:

        if properties.reactions is None:
            return

        if properties.type.is_normal or properties.type.is_headline:
            self._log.warning(
                "Reaction with undefined message type %s", properties.type
            )
            raise NodeProcessed

        if properties.type.is_error:
            # TODO: Maybe inform the user and delete
            # the reactions from the database
            # But this handler here is only executed if the
            # server adds the message content to the error
            raise NodeProcessed

        assert properties.jid is not None
        if properties.type.is_groupchat and properties.jid.is_bare:
            # Reactions from the bare groupchat jid are not defined
            raise NodeProcessed

        remote_jid = properties.remote_jid
        assert remote_jid is not None

        timestamp = get_message_timestamp(properties)

        muc_data = None
        if properties.type.is_groupchat:
            muc_data = self._client.get_module("MUC").get_muc_data(remote_jid)
            if muc_data is None:
                self._log.warning("Reaction message from unknown MUC: %s", remote_jid)
                raise NodeProcessed

        own_bare_jid = self._get_own_bare_jid()

        m_type, direction = get_chat_type_and_direction(
            muc_data, own_bare_jid, properties
        )

        occupant = None
        occupant_id = None
        if m_type in (MessageType.GROUPCHAT, MessageType.PM):
            assert properties.jid is not None
            contact = self._client.get_module("Contacts").get_contact(
                properties.jid, groupchat=True
            )

            assert isinstance(contact, GroupchatParticipant)
            occupant = get_occupant_info(
                account=self._account,
                remote_jid=remote_jid,
                own_bare_jid=own_bare_jid,
                direction=direction,
                m_type=m_type,
                timestamp=timestamp,
                contact=contact,
                properties=properties,
            )

            if occupant is None:
                self._log.info("Reactions not supported without occupant-id")
                raise NodeProcessed

            occupant_id = occupant.id

        if not properties.reactions.emojis:
            app.storage.archive.delete_reaction(
                account=self._account,
                jid=remote_jid,
                occupant_id=occupant_id,
                reaction_id=properties.reactions.id,
                direction=direction,
            )
            self._log.info("Delete reactions: %s", properties.jid)
            app.ged.raise_event(
                ReactionUpdated(
                    account=self._account,
                    jid=remote_jid,
                    id=properties.reactions.id,
                    direction=direction,
                    occupant=occupant,
                    emojis=None,
                    message=None,
                    is_mam_message=properties.is_mam_message,
                )
            )
            raise NodeProcessed

        # Set arbitrary limit of max reactions to prevent
        # performance problems when loading and displaying them.
        # Check if reactions qualify as emojis.
        valid, invalid = normalize_reactions(list(properties.reactions.emojis))
        if invalid:
            codepoints = ", ".join([convert_to_codepoints(i) for i in invalid])
            self._log.warning("Reactions did not qualify as emoji: %s", codepoints)

        if not valid:
            raise NodeProcessed

        reaction = mod.Reaction(
            account_=self._account,
            remote_jid_=remote_jid,
            occupant_=occupant,
            id=properties.reactions.id,
            direction=direction,
            emojis=";".join(valid),
            timestamp=timestamp,
        )

        self._log.info("Received reactions: %s", reaction)
        app.storage.archive.upsert_row2(reaction, return_full=False)

        id_type = (
            "stanza-id"
            if m_type in (MessageType.GROUPCHAT, MessageType.PM)
            else "message-id"
        )

        message = app.storage.archive.get_message_with_id(
            self._account,
            remote_jid,
            id_type,
            properties.reactions.id,
            default_options=False,
        )

        app.ged.raise_event(
            ReactionUpdated(
                account=self._account,
                jid=remote_jid,
                id=properties.reactions.id,
                direction=direction,
                occupant=occupant,
                emojis=valid,
                message=message,
                is_mam_message=properties.is_mam_message,
            )
        )

        raise NodeProcessed

    def send_reaction(
        self,
        contact: types.ChatContactT,
        reaction_id: str,
        reactions: set[str],
    ) -> None:

        valid, invalid = normalize_reactions(list(reactions))
        if invalid:
            codepoints = ", ".join([convert_to_codepoints(i) for i in invalid])
            self._log.warning("Trying to send invalid reactions: %s", codepoints)

        message = OutgoingMessage(
            account=self._account,
            contact=contact,
            reaction_data=(reaction_id, valid),
            play_sound=False,
        )

        self._client.send_message(message)
        self._log.info("Send %s: %s", reactions, contact.jid)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Any
from typing import cast
from typing import Literal
from typing import NamedTuple
from typing import overload
from typing import TypedDict

import inspect
import json
import logging
import sqlite3
import sys
import uuid
import weakref
from collections import defaultdict
from collections import namedtuple
from collections.abc import Callable
from pathlib import Path

from gi.repository import GLib
from nbxmpp.protocol import JID

from gajim import IS_PORTABLE
from gajim.common import app
from gajim.common import configpaths
from gajim.common.helpers import deep_update
from gajim.common.setting_values import ACCOUNT_SETTINGS
from gajim.common.setting_values import AllAccountSettings
from gajim.common.setting_values import AllAccountSettingsT
from gajim.common.setting_values import AllContactSettings
from gajim.common.setting_values import AllContactSettingsT
from gajim.common.setting_values import AllGroupChatSettings
from gajim.common.setting_values import AllGroupChatSettingsT
from gajim.common.setting_values import AllSettingsT
from gajim.common.setting_values import AllWorkspaceSettings
from gajim.common.setting_values import AllWorkspaceSettingsT
from gajim.common.setting_values import APP_SETTINGS
from gajim.common.setting_values import BoolAccountSettings
from gajim.common.setting_values import BoolContactSettings
from gajim.common.setting_values import BoolGroupChatSettings
from gajim.common.setting_values import BoolSettings
from gajim.common.setting_values import DEFAULT_SOUNDEVENT_SETTINGS
from gajim.common.setting_values import FloatSettings
from gajim.common.setting_values import HAS_ACCOUNT_DEFAULT
from gajim.common.setting_values import HAS_APP_DEFAULT
from gajim.common.setting_values import INITAL_WORKSPACE
from gajim.common.setting_values import IntAccountSettings
from gajim.common.setting_values import IntGroupChatSettings
from gajim.common.setting_values import IntSettings
from gajim.common.setting_values import OpenChatsSettingT
from gajim.common.setting_values import PLUGIN_SETTINGS
from gajim.common.setting_values import PROXY_EXAMPLES
from gajim.common.setting_values import PROXY_SETTINGS
from gajim.common.setting_values import StringAccountSettings
from gajim.common.setting_values import StringContactSettings
from gajim.common.setting_values import StringGroupChatSettings
from gajim.common.setting_values import StringSettings
from gajim.common.setting_values import StringWorkspaceSettings
from gajim.common.setting_values import WORKSPACE_SETTINGS
from gajim.common.setting_values import WorkspaceSettings
from gajim.common.storage.base import Encoder
from gajim.common.storage.base import json_decoder
from gajim.common.util.classes import SettingsAction

SETTING_TYPE = bool | int | str | object


log = logging.getLogger("gajim.c.settings")

CURRENT_USER_VERSION = 7

CREATE_SQL = """
    CREATE TABLE settings (
            name TEXT UNIQUE,
            settings TEXT
    );

    CREATE TABLE account_settings (
            account TEXT UNIQUE,
            settings TEXT
    );

    INSERT INTO settings(name, settings) VALUES ('app', '{{}}');
    INSERT INTO settings(name, settings) VALUES ('soundevents', '{{}}');
    INSERT INTO settings(name, settings) VALUES ('proxies', '{proxies}');
    INSERT INTO settings(name, settings) VALUES ('plugins', '{{}}');
    INSERT INTO settings(name, settings) VALUES ('workspaces', '{workspaces}');
    INSERT INTO settings(name, settings) VALUES ('window_sizes', '{{}}');

    PRAGMA user_version={version};
    """.format(  # noqa: UP032
    proxies=json.dumps(PROXY_EXAMPLES),
    workspaces=json.dumps(INITAL_WORKSPACE),
    version=CURRENT_USER_VERSION,
)


_SignalCallable = Callable[[Any, str, str | None, JID | None], Any]
_CallbackDict = dict[
    tuple[str, str | None, JID | None], list[weakref.WeakMethod[_SignalCallable]]
]

if app.is_flatpak():
    app_overrides = "/app/app-overrides.json"
else:
    app_overrides = "/etc/gajim/app-overrides.json"
OVERRIDES_PATH = Path(app_overrides)


class SettingsDictT(TypedDict):
    app: dict[str, Any]
    plugins: dict[str, dict[str, Any]]
    workspaces: dict[str, dict[str, WorkspaceSettings]]
    soundevents: dict[str, dict[str, Any]]
    proxies: dict[str, dict[str, Any]]
    window_sizes: dict[str, tuple[int, int]]


class Settings:
    def __init__(self, in_memory: bool = False):
        self._con = cast(sqlite3.Connection, None)
        self._commit_scheduled = None
        self._in_memory = in_memory

        self._settings: SettingsDictT = {}
        self._app_overrides: dict[str, AllSettingsT] = {}
        self._account_settings: dict[str, Any | dict[str, dict[JID | str, Any]]] = {}

        self._callbacks: _CallbackDict = defaultdict(list)

    def connect_signal(
        self,
        setting: str,
        func: _SignalCallable,
        account: str | None = None,
        jid: JID | None = None,
    ) -> None:
        if not inspect.ismethod(func):
            # static methods are not bound to an object so we can’t easily
            # remove the func once it should not be called anymore
            raise ValueError("Only bound methods can be connected")

        weak_func = weakref.WeakMethod(func)
        self._callbacks[(setting, account, jid)].append(weak_func)

    def disconnect_signals(self, object_: object) -> Any:
        for handlers in self._callbacks.values():
            for handler in list(handlers):
                if isinstance(handler, tuple):
                    widget = handler[1]
                    if widget is object_:
                        handlers.remove(handler)
                else:
                    func = handler()
                    if func is None or func.__self__ is object_:
                        handlers.remove(handler)

    def create_action(self, setting: BoolSettings) -> SettingsAction:
        # Currently only boolean app settings are supported
        value = self.get_app_setting(setting)
        action = SettingsAction(
            name=setting.replace("_", "-"),
            parameter_type=GLib.VariantType("b"),
            state=GLib.Variant("b", value),
        )

        self.bind_signal(setting, action, "change_state")
        return action

    def bind_signal(
        self,
        setting: str,
        widget: Any,
        func_name: str,
        account: str | None = None,
        jid: JID | None = None,
        inverted: bool = False,
        default_text: str | None = None,
    ) -> None:

        callbacks = self._callbacks[(setting, account, jid)]
        func = getattr(widget, func_name)
        callbacks.append((func, widget, inverted, default_text))

    def _notify(
        self,
        value: Any,
        setting: str,
        account: str | None = None,
        jid: JID | None = None,
    ) -> None:

        log.info("Signal: %s changed", setting)

        callbacks = self._callbacks[(setting, account, jid)]
        for func in list(callbacks):
            if isinstance(func, tuple):
                func, _widget, inverted, default_text = func
                if isinstance(value, bool) and inverted:
                    value = not value

                if value == "" and default_text is not None:
                    value = default_text

                try:
                    func(value)
                except Exception:
                    log.exception("Error while executing signal callback")
                continue

            if func() is None:
                callbacks.remove(func)
                continue

            func = func()
            if func is None:
                continue

            try:
                func(value, setting, account, jid)
            except Exception:
                log.exception("Error while executing signal callback")

    def init(self, init_config_path: Path | None = None) -> None:
        self._setup_installation_defaults()
        if self._in_memory:
            self._connect_in_memory_database()
        else:
            self._connect_database()
        self._load_settings()
        self._load_account_settings()
        self._migrate_database()
        self._load_settings_from_path(init_config_path)
        self._load_app_overrides()
        self._commit()

    @staticmethod
    def _setup_installation_defaults() -> None:
        if IS_PORTABLE:
            APP_SETTINGS["use_keyring"] = False

        if sys.platform == "win32":
            APP_SETTINGS["app_font_size"] = 1.125

    def _load_settings_from_path(self, path: Path | None) -> None:
        if path is None:
            return

        log.info("Load settings from path: %s", path)

        try:
            settings = json.loads(path.read_bytes(), object_hook=json_decoder)
        except Exception as error:
            sys.exit(f"Unable to load settings from path: {error}")

        for cat in self._settings:
            self._settings[cat] = deep_update(
                self._settings[cat], settings.get(cat, {})
            )

        account_settings = settings.get("accounts", {})
        for account in account_settings:
            if account not in self._account_settings:
                self.add_account(account)

        self._account_settings = deep_update(
            self._account_settings, settings.get("accounts", {})
        )

        self._commit_all()

    def _load_app_overrides(self) -> None:
        if not OVERRIDES_PATH.exists():
            return

        with OVERRIDES_PATH.open(encoding="utf8") as f:
            try:
                self._app_overrides = json.load(f)
            except Exception:
                log.exception("Failed to load overrides")
                return

        self._settings["app"].update(self._app_overrides)

    def export_to_json(self, path: Path) -> None:
        config = self._settings.copy()
        config["accounts"] = self._account_settings
        json_string = json.dumps(config, cls=Encoder, indent=2)
        path.write_text(json_string)

    @staticmethod
    def _namedtuple_factory(cursor: sqlite3.Cursor, row: tuple[Any, ...]) -> NamedTuple:
        fields = [col[0] for col in cursor.description]
        return namedtuple("Row", fields)(*row)  # type: ignore

    def _connect_database(self) -> None:
        path = configpaths.get("SETTINGS")
        if path.is_dir():
            log.error("%s is a directory but should be a file", path)
            sys.exit()

        if not path.exists():
            self._create_database(CREATE_SQL, path)

        self._con = sqlite3.connect(path)
        self._con.row_factory = self._namedtuple_factory

    def _connect_in_memory_database(self) -> None:
        log.info("Creating in memory")
        self._con = sqlite3.connect(":memory:")
        self._con.row_factory = self._namedtuple_factory

        try:
            self._con.executescript(CREATE_SQL)
        except Exception:
            log.exception("Error")
            self._con.close()
            sys.exit()

        self._con.commit()

    @staticmethod
    def _create_database(statement: str, path: Path) -> None:
        log.info("Creating %s", path)
        con = sqlite3.connect(path)

        try:
            con.executescript(statement)
        except Exception:
            log.exception("Error")
            con.close()
            path.unlink()
            sys.exit()

        con.commit()
        con.close()
        path.chmod(0o600)

    def _get_user_version(self) -> int:
        return self._con.execute("PRAGMA user_version").fetchone()[0]

    def _set_user_version(self, version: int) -> None:
        self._con.execute(f"PRAGMA user_version = {version}")
        self._commit()

    def _commit(self, schedule: bool = False) -> None:
        if not schedule:
            if self._commit_scheduled is not None:
                GLib.source_remove(self._commit_scheduled)
                self._commit_scheduled = None
            log.info("Commit")
            self._con.commit()

        elif self._commit_scheduled is None:
            self._commit_scheduled = GLib.timeout_add(200, self._scheduled_commit)

    def _commit_all(self) -> None:
        for account in self._account_settings:
            self._commit_account_settings(account)

        for key in self._settings:
            self._commit_settings(key)

    def save(self) -> None:
        self._commit()

    def _scheduled_commit(self) -> None:
        self._commit_scheduled = None
        log.info("Commit")
        self._con.commit()

    def _migrate_database(self) -> None:
        if self._in_memory:
            return

        try:
            self._migrate()
        except Exception:
            self._con.close()
            log.exception("Error")
            sys.exit()

    def _migrate(self) -> None:
        version = self._get_user_version()
        if version < 1:
            sql = """INSERT INTO settings(name, settings)
                     VALUES ('workspaces', ?)"""
            self._con.execute(sql, (json.dumps(INITAL_WORKSPACE),))
            self._settings["workspaces"] = INITAL_WORKSPACE
            self._commit_settings("workspaces")
            self._set_user_version(1)

        if version < 2:
            # Migrate open chats to new key and format
            for workspace in self._settings["workspaces"].values():
                open_chats: list[dict[str, Any]] = []
                for open_chat in workspace.get("open_chats", []):
                    account, jid, type_, pinned = open_chat
                    open_chats.append(
                        {
                            "account": account,
                            "jid": jid,
                            "type": type_,
                            "pinned": pinned,
                            "position": -1,
                        }
                    )

                workspace["chats"] = open_chats
                workspace.pop("open_chats", None)

            self._commit_settings("workspaces")
            self._set_user_version(2)

        if version < 3:
            # Migrate open chats to new key and format
            for account_settings in self._account_settings.values():
                if account_settings["account"].get("active") is None:
                    account_settings["account"]["active"] = True

            for account in self._account_settings:
                self._commit_account_settings(account)

            self._set_user_version(3)

        if version < 4:
            value = self._settings["app"].get("chat_timestamp_format")
            if value is not None:
                self._settings["app"]["time_format"] = value

            value = self._settings["app"].get("date_timestamp_format")
            if value is not None:
                self._settings["app"]["date_format"] = value

            self._commit_settings("app")
            self._set_user_version(4)

        if version < 5:
            self._settings["app"].pop("muclumbus_api_http_uri", None)
            self._commit_settings("app")
            self._set_user_version(5)

        if version < 6:
            for account_settings in self._account_settings.values():
                settings = account_settings["account"]
                localpart = settings.get("name", "")
                domain = settings.get("hostname", "")
                try:
                    address = JID.from_string(f"{localpart}@{domain}")
                except Exception as error:
                    log.warning("Unable to migrate address: %s", error)
                    address = JID.from_string("unknown@unknown.invalid")

                settings["address"] = str(address)

            for account in self._account_settings:
                self._commit_account_settings(account)

            self._set_user_version(6)

        if version < 7:
            sql = """INSERT INTO settings(name, settings)
                     VALUES ('window_sizes', '{}')"""
            self._con.execute(sql)
            self._settings["window_sizes"] = {}
            self._set_user_version(7)

    def close(self) -> None:
        log.info("Close settings")
        self._con.commit()
        self._con.close()
        self._con = cast(sqlite3.Connection, None)

    def _load_settings(self) -> None:
        settings = self._con.execute("SELECT * FROM settings").fetchall()
        for row in settings:
            log.info("Load %s settings", row.name)
            self._settings[row.name] = json.loads(
                row.settings, object_hook=json_decoder
            )

    def _load_account_settings(self) -> None:
        account_settings = self._con.execute(
            "SELECT * FROM account_settings"
        ).fetchall()
        for row in account_settings:
            log.info("Load account settings: %s", row.account)
            self._account_settings[row.account] = json.loads(
                row.settings, object_hook=json_decoder
            )

    def _commit_account_settings(self, account: str, schedule: bool = True) -> None:
        log.info("Set account settings: %s", account)
        self._con.execute(
            "UPDATE account_settings SET settings = ? WHERE account = ?",
            (json.dumps(self._account_settings[account], cls=Encoder), account),
        )

        self._commit(schedule=schedule)

    def _commit_settings(self, name: str, schedule: bool = True) -> None:
        log.info("Set settings: %s", name)
        self._con.execute(
            "UPDATE settings SET settings = ? WHERE name = ?",
            (json.dumps(self._settings[name], cls=Encoder), name),
        )

        self._commit(schedule=schedule)

    def has_app_override(self, setting: str) -> bool:
        return setting in self._app_overrides

    @overload
    def get_app_setting(self, setting: BoolSettings) -> bool: ...

    @overload
    def get_app_setting(self, setting: StringSettings) -> str: ...

    @overload
    def get_app_setting(self, setting: IntSettings) -> int: ...

    @overload
    def get_app_setting(self, setting: FloatSettings) -> float: ...

    def get_app_setting(self, setting: str) -> AllSettingsT:
        if setting not in APP_SETTINGS:
            raise ValueError(f"Invalid app setting: {setting}")

        try:
            return self._settings["app"][setting]
        except KeyError:
            return APP_SETTINGS[setting]

    get = get_app_setting

    @overload
    def set_app_setting(self, setting: BoolSettings, value: bool | None) -> None: ...

    @overload
    def set_app_setting(self, setting: StringSettings, value: str | None) -> None: ...

    @overload
    def set_app_setting(self, setting: IntSettings, value: int | None) -> None: ...

    @overload
    def set_app_setting(self, setting: FloatSettings, value: float | None) -> None: ...

    @overload
    def set_app_setting(
        self, setting: Literal["workspace_order"], value: list[str]
    ) -> None: ...

    def set_app_setting(self, setting: str, value: AllSettingsT | None) -> None:

        if setting not in APP_SETTINGS:
            raise ValueError(f"Invalid app setting: {setting}")

        if setting in self._app_overrides:
            log.warning(
                "Changing %s is not allowed because there exists an override", setting
            )
            return

        default = APP_SETTINGS[setting]
        if not isinstance(value, type(default)) and value is not None:
            raise TypeError(f"Invalid type for {setting}: {value} {type(value)}")

        if value is None:
            try:
                del self._settings["app"][setting]
            except KeyError:
                pass

            self._commit_settings("app")
            self._notify(default, setting)
            return

        self._settings["app"][setting] = value

        self._commit_settings("app")
        self._notify(value, setting)

    set = set_app_setting

    def set_window_size(self, window_name: str, width: int, height: int) -> None:
        self._settings["window_sizes"][window_name] = (width, height)
        self._commit_settings("window_sizes")

    def get_window_size(self, window_name: str) -> tuple[int, int] | None:
        return self._settings["window_sizes"].get(window_name)

    def get_plugin_setting(self, plugin: str, setting: str) -> SETTING_TYPE:
        if setting not in PLUGIN_SETTINGS:
            raise ValueError(f"Invalid plugin setting: {setting}")

        if plugin not in self._settings["plugins"]:
            raise ValueError(f"Unknown plugin {plugin}")

        try:
            return self._settings["plugins"][plugin][setting]
        except KeyError:
            return PLUGIN_SETTINGS[setting]

    def get_plugins(self) -> list[str]:
        return list(self._settings["plugins"].keys())

    def set_plugin_setting(self, plugin: str, setting: str, value: bool) -> None:

        if setting not in PLUGIN_SETTINGS:
            raise ValueError(f"Invalid plugin setting: {setting}")

        default = PLUGIN_SETTINGS[setting]
        if not isinstance(value, type(default)):
            raise TypeError(f"Invalid type for {setting}: {value} {type(value)}")

        if plugin in self._settings["plugins"]:
            self._settings["plugins"][plugin][setting] = value
        else:
            self._settings["plugins"][plugin] = {setting: value}

        self._commit_settings("plugins")

    def remove_plugin(self, plugin: str) -> None:
        try:
            del self._settings["plugins"][plugin]
        except KeyError:
            pass

    def add_account(self, account: str) -> None:
        log.info("Add account: %s", account)
        if account in self._account_settings:
            raise ValueError("Account %s exists already" % account)

        self._account_settings[account] = {
            "account": {},
            "contact": {},
            "group_chat": {},
        }
        self._con.execute(
            "INSERT INTO account_settings(account, settings) VALUES(?, ?)",
            (account, json.dumps(self._account_settings[account])),
        )
        self._commit()

    def remove_account(self, account: str) -> None:
        if account not in self._account_settings:
            raise ValueError(f"Unknown account: {account}")

        del self._account_settings[account]
        self._con.execute("DELETE FROM account_settings WHERE account = ?", (account,))
        self._commit()

    def get_accounts(self) -> list[str]:
        return list(self._account_settings.keys())

    def account_exists(self, jid: str) -> bool:
        for account in self._account_settings:
            address = self.get_account_setting(account, "address")
            if jid == JID.from_string(address):
                return True
        return False

    def get_active_accounts(self) -> list[str]:
        active: list[str] = []
        for account in self._account_settings:
            if self.get_account_setting(account, "active"):
                active.append(account)
        return active

    def get_account_from_jid(self, jid: JID) -> str:
        for account in self._account_settings:
            address = self.get_account_setting(account, "address")
            if jid == JID.from_string(address):
                return account

        raise ValueError(f"No account found for: {jid}")

    @overload
    def get_account_setting(
        self, account: str, setting: StringAccountSettings
    ) -> str: ...

    @overload
    def get_account_setting(self, account: str, setting: IntAccountSettings) -> int: ...

    @overload
    def get_account_setting(
        self, account: str, setting: BoolAccountSettings
    ) -> bool: ...

    def get_account_setting(
        self, account: str, setting: AllAccountSettings
    ) -> AllAccountSettingsT:

        if account not in self._account_settings:
            raise ValueError(f"Account missing: {account}")

        if setting not in ACCOUNT_SETTINGS["account"]:
            raise ValueError(f"Invalid account setting: {setting}")

        try:
            return self._account_settings[account]["account"][setting]
        except KeyError:
            return ACCOUNT_SETTINGS["account"][setting]

    @overload
    def set_account_setting(
        self, account: str, setting: StringAccountSettings, value: str | None
    ) -> None: ...

    @overload
    def set_account_setting(
        self, account: str, setting: IntAccountSettings, value: int | None
    ) -> None: ...

    @overload
    def set_account_setting(
        self, account: str, setting: BoolAccountSettings, value: bool | None
    ) -> None: ...

    def set_account_setting(
        self,
        account: str,
        setting: AllAccountSettings,
        value: AllAccountSettingsT | None,
    ) -> None:

        if account not in self._account_settings:
            raise ValueError(f"Account missing: {account}")

        if setting not in ACCOUNT_SETTINGS["account"]:
            raise ValueError(f"Invalid account setting: {setting}")

        default = ACCOUNT_SETTINGS["account"][setting]
        if not isinstance(value, type(default)) and value is not None:
            raise TypeError(f"Invalid type for {setting}: {value} {type(value)}")

        if value is None:
            try:
                del self._account_settings[account]["account"][setting]
            except KeyError:
                pass

            self._commit_account_settings(account)
            self._notify(default, setting, account)
            return

        self._account_settings[account]["account"][setting] = value

        self._commit_account_settings(account)
        self._notify(value, setting, account)

    @overload
    def get_group_chat_setting(
        self, account: str, jid: JID, setting: IntGroupChatSettings
    ) -> int: ...

    @overload
    def get_group_chat_setting(
        self, account: str, jid: JID, setting: BoolGroupChatSettings
    ) -> bool: ...

    @overload
    def get_group_chat_setting(
        self, account: str, jid: JID, setting: StringGroupChatSettings
    ) -> str: ...

    def get_group_chat_setting(
        self, account: str, jid: JID, setting: AllGroupChatSettings
    ) -> AllGroupChatSettingsT:

        if account not in self._account_settings:
            raise ValueError(f"Account missing: {account}")

        if setting not in ACCOUNT_SETTINGS["group_chat"]:
            raise ValueError(f"Invalid group chat setting: {setting}")

        try:
            return self._account_settings[account]["group_chat"][jid][setting]
        except KeyError:
            client = app.get_client(account)
            contact = client.get_module("Contacts").get_contact(jid)
            context = contact.muc_context
            if context is None:
                # If there is no disco info available
                # to determine the context assume public
                log.warning("Unable to determine context for: %s", jid)
                context = "public"

            default = ACCOUNT_SETTINGS["group_chat"][setting]
            if default is HAS_APP_DEFAULT:
                context_default_setting = f"gc_{setting}_{context}_default"
                if context_default_setting in APP_SETTINGS:
                    return self.get_app_setting(context_default_setting)
                return self.get_app_setting(f"gc_{setting}_default")

            if default is HAS_ACCOUNT_DEFAULT:
                default_settings = [
                    f"gc_{setting}_{context}_default",
                    f"gc_{setting}_default",
                    f"{setting}_default",
                ]

                for default_setting in default_settings:
                    if default_setting in ACCOUNT_SETTINGS["account"]:
                        return self.get_account_setting(account, default_setting)

                raise ValueError(f"No default setting found for {setting}")

            return default

    @overload
    def set_group_chat_setting(
        self,
        account: str,
        jid: JID,
        setting: StringGroupChatSettings,
        value: str | None,
    ) -> None: ...

    @overload
    def set_group_chat_setting(
        self, account: str, jid: JID, setting: IntGroupChatSettings, value: int | None
    ) -> None: ...

    @overload
    def set_group_chat_setting(
        self, account: str, jid: JID, setting: BoolGroupChatSettings, value: bool | None
    ) -> None: ...

    def set_group_chat_setting(
        self,
        account: str,
        jid: JID,
        setting: AllGroupChatSettings,
        value: AllGroupChatSettingsT | None,
    ) -> None:

        if account not in self._account_settings:
            raise ValueError(f"Account missing: {account}")

        if setting not in ACCOUNT_SETTINGS["group_chat"]:
            raise ValueError(f"Invalid group chat setting: {setting}")

        default = ACCOUNT_SETTINGS["group_chat"][setting]
        if default in (HAS_APP_DEFAULT, HAS_ACCOUNT_DEFAULT):
            context = "public"
            if app.account_is_connected(account):
                client = app.get_client(account)
                contact = client.get_module("Contacts").get_contact(jid, groupchat=True)
                context = contact.muc_context
                if context is None:
                    # If there is no disco info available
                    # to determine the context assume public
                    log.warning("Unable to determine context for: %s", jid)
                    context = "public"

            default_store = APP_SETTINGS
            if default is HAS_ACCOUNT_DEFAULT:
                default_store = ACCOUNT_SETTINGS["account"]

            default_settings = [
                f"gc_{setting}_{context}_default",
                f"gc_{setting}_default",
                f"{setting}_default",
            ]

            default = None
            for default_setting in default_settings:
                if default_setting in default_store:
                    default = default_store.get(default_setting)
                    break

            if default is None:
                raise ValueError(f"No default setting found for {setting}")

        if not isinstance(value, type(default)) and value is not None:
            raise TypeError(f"Invalid type for {setting}: {value} {type(value)}")

        if value is None:
            try:
                del self._account_settings[account]["group_chat"][jid][setting]
            except KeyError:
                pass

            self._commit_account_settings(account)
            self._notify(default, setting, account, jid)
            return

        group_chat_settings = self._account_settings[account]["group_chat"]
        if jid not in group_chat_settings:
            group_chat_settings[str(jid)] = {setting: value}
        else:
            group_chat_settings[jid][setting] = value

        self._commit_account_settings(account)
        self._notify(value, setting, account, jid)

    def set_group_chat_settings(
        self,
        setting: str,
        account: str | None,
        value: SETTING_TYPE,
        context: str | None = None,
    ) -> None:

        if account is None:
            settings = self._account_settings.items()
        else:
            settings = [(account, self._account_settings[account])]

        for acc, acc_settings in settings:
            for jid in acc_settings["group_chat"]:
                if context is not None:
                    client = app.get_client(acc)
                    contact = client.get_module("Contacts").get_contact(
                        jid, groupchat=True
                    )
                    if contact.muc_context != context:
                        continue
                self.set_group_chat_setting(acc, jid, setting, value)

    @overload
    def get_contact_setting(
        self, account: str, jid: JID, setting: BoolContactSettings
    ) -> bool: ...

    @overload
    def get_contact_setting(
        self, account: str, jid: JID, setting: StringContactSettings
    ) -> str: ...

    def get_contact_setting(
        self, account: str, jid: JID, setting: AllContactSettings
    ) -> AllContactSettingsT:

        if account not in self._account_settings:
            raise ValueError(f"Account missing: {account}")

        if setting not in ACCOUNT_SETTINGS["contact"]:
            raise ValueError(f"Invalid contact setting: {setting}")

        try:
            return self._account_settings[account]["contact"][jid][setting]
        except KeyError:
            default = ACCOUNT_SETTINGS["contact"][setting]
            if default is HAS_APP_DEFAULT:
                return self.get_app_setting(f"{setting}_default")

            if default is HAS_ACCOUNT_DEFAULT:
                return self.get_account_setting(account, f"{setting}_default")

            return default

    @overload
    def set_contact_setting(
        self, account: str, jid: JID, setting: StringContactSettings, value: str | None
    ) -> None: ...

    @overload
    def set_contact_setting(
        self, account: str, jid: JID, setting: BoolContactSettings, value: bool | None
    ) -> None: ...

    def set_contact_setting(
        self,
        account: str,
        jid: JID,
        setting: AllContactSettings,
        value: AllContactSettingsT | None,
    ) -> None:

        if account not in self._account_settings:
            raise ValueError(f"Account missing: {account}")

        if setting not in ACCOUNT_SETTINGS["contact"]:
            raise ValueError(f"Invalid contact setting: {setting}")

        default = ACCOUNT_SETTINGS["contact"][setting]
        if default in (HAS_APP_DEFAULT, HAS_ACCOUNT_DEFAULT):
            default_store = APP_SETTINGS
            if default is HAS_ACCOUNT_DEFAULT:
                default_store = ACCOUNT_SETTINGS["account"]

            default = default_store[f"{setting}_default"]

        if not isinstance(value, type(default)) and value is not None:
            raise TypeError(f"Invalid type for {setting}: {value} {type(value)}")

        if value is None:
            try:
                del self._account_settings[account]["contact"][jid][setting]
            except KeyError:
                pass

            self._commit_account_settings(account)
            self._notify(default, setting, account, jid)
            return

        contact_settings = self._account_settings[account]["contact"]
        if jid not in contact_settings:
            contact_settings[str(jid)] = {setting: value}
        else:
            contact_settings[jid][setting] = value

        self._commit_account_settings(account)
        self._notify(value, setting, account, jid)

    def set_contact_settings(
        self, setting: str, account: str | None, value: SETTING_TYPE
    ) -> None:

        if account is None:
            settings = self._account_settings.items()
        else:
            settings = [(account, self._account_settings[account])]

        for acc, acc_settings in settings:
            for jid in acc_settings["contact"]:
                self.set_contact_setting(acc, jid, setting, value)

    def set_soundevent_setting(
        self, event_name: str, setting: str, value: SETTING_TYPE
    ) -> None:

        if event_name not in DEFAULT_SOUNDEVENT_SETTINGS:
            raise ValueError(f"Invalid soundevent: {event_name}")

        if setting not in DEFAULT_SOUNDEVENT_SETTINGS[event_name]:
            raise ValueError(f"Invalid soundevent setting: {setting}")

        default = DEFAULT_SOUNDEVENT_SETTINGS[event_name][setting]
        if not isinstance(value, type(default)):
            raise TypeError(f"Invalid type for {setting}: {value} {type(value)}")

        if event_name not in self._settings["soundevents"]:
            self._settings["soundevents"][event_name] = {setting: value}
        else:
            self._settings["soundevents"][event_name][setting] = value

        self._commit_settings("soundevents")

    def get_soundevent_settings(self, event_name: str) -> dict[str, SETTING_TYPE]:

        if event_name not in DEFAULT_SOUNDEVENT_SETTINGS:
            raise ValueError(f"Invalid soundevent: {event_name}")

        settings = DEFAULT_SOUNDEVENT_SETTINGS[event_name].copy()
        user_settings = self._settings["soundevents"].get(event_name, {})
        settings.update(user_settings)
        return settings

    def set_proxy_setting(
        self, proxy_name: str, setting: str, value: SETTING_TYPE
    ) -> None:

        if setting not in PROXY_SETTINGS:
            raise ValueError(f"Invalid proxy setting: {setting}")

        default = PROXY_SETTINGS[setting]
        if not isinstance(value, type(default)):
            raise TypeError(f"Invalid type for {setting}: {value} {type(value)}")

        if proxy_name in self._settings["proxies"]:
            self._settings["proxies"][proxy_name][setting] = value
        else:
            self._settings["proxies"][proxy_name] = {setting: value}

        self._commit_settings("proxies")

    def get_proxy_settings(self, proxy_name: str) -> dict[str, SETTING_TYPE]:
        if proxy_name not in self._settings["proxies"]:
            raise ValueError(f"Unknown proxy: {proxy_name}")

        settings = PROXY_SETTINGS.copy()
        user_settings = self._settings["proxies"][proxy_name]
        settings.update(user_settings)
        return settings

    def get_proxies(self) -> list[str]:
        return list(self._settings["proxies"].keys())

    def add_proxy(self, proxy_name: str) -> None:
        if proxy_name in self._settings["proxies"]:
            raise ValueError(f"Proxy already exists: {proxy_name}")

        self._settings["proxies"][proxy_name] = {}

    def rename_proxy(self, old_proxy_name: str, new_proxy_name: str) -> None:
        settings = self._settings["proxies"].pop(old_proxy_name)
        self._settings["proxies"][new_proxy_name] = settings

    def remove_proxy(self, proxy_name: str) -> None:
        if proxy_name not in self._settings["proxies"]:
            raise ValueError(f"Unknown proxy: {proxy_name}")

        del self._settings["proxies"][proxy_name]
        self._commit_settings("proxies")

        if self.get_app_setting("global_proxy") == proxy_name:
            self.set_app_setting("global_proxy", None)

        for account in self._account_settings:
            if self.get_account_setting(account, "proxy") == proxy_name:
                self.set_account_setting(account, "proxy", None)

    @overload
    def set_workspace_setting(
        self, workspace_id: str, setting: StringWorkspaceSettings, value: str
    ) -> None: ...

    @overload
    def set_workspace_setting(
        self, workspace_id: str, setting: Literal["chats"], value: OpenChatsSettingT
    ) -> None: ...

    def set_workspace_setting(
        self,
        workspace_id: str,
        setting: AllWorkspaceSettings,
        value: AllWorkspaceSettingsT,
    ) -> None:

        if setting not in WORKSPACE_SETTINGS:
            raise ValueError(f"Invalid workspace setting: {setting}")

        if workspace_id not in self._settings["workspaces"]:
            raise ValueError(f"Workspace does not exists: {workspace_id}")

        default = WORKSPACE_SETTINGS[setting]
        if not isinstance(value, type(default)):
            raise TypeError(f"Invalid type for {setting}: {value} {type(value)}")

        self._settings["workspaces"][workspace_id][setting] = value
        self._commit_settings("workspaces")

    @overload
    def get_workspace_setting(
        self, workspace_id: str, setting: Literal["chats"]
    ) -> OpenChatsSettingT: ...

    @overload
    def get_workspace_setting(
        self, workspace_id: str, setting: StringWorkspaceSettings
    ) -> str: ...

    def get_workspace_setting(
        self, workspace_id: str, setting: AllWorkspaceSettings
    ) -> AllWorkspaceSettingsT:

        if setting not in WORKSPACE_SETTINGS:
            raise ValueError(f"Invalid workspace setting: {setting}")

        if workspace_id not in self._settings["workspaces"]:
            raise ValueError(f"Workspace does not exists: {workspace_id}")

        try:
            return self._settings["workspaces"][workspace_id][setting]
        except KeyError:
            return WORKSPACE_SETTINGS[setting]

    def get_workspace_count(self) -> int:
        return len(self._settings["workspaces"].keys())

    def get_workspaces(self) -> list[str]:
        workspace_order = app.settings.get_app_setting("workspace_order")

        def sort_workspaces(workspace_id: str) -> int:
            try:
                return workspace_order.index(workspace_id)
            except ValueError:
                # Handle the case that a workspace id is for some reason not
                # in the workspace order list
                return 10000

        workspaces = list(self._settings["workspaces"].keys())
        workspaces.sort(key=sort_workspaces)
        return workspaces

    def workspace_exists(self, workspace_id: str) -> bool:
        return workspace_id in app.settings.get_workspaces()

    def add_workspace(self, name: str) -> str:
        id_ = str(uuid.uuid4())
        self._settings["workspaces"][id_] = {
            "name": name,
        }
        self._commit_settings("workspaces")
        return id_

    def remove_workspace(self, id_: str) -> None:
        del self._settings["workspaces"][id_]
        self._commit_settings("workspaces")

    def shutdown(self) -> None:
        if self._commit_scheduled is not None:
            GLib.source_remove(self._commit_scheduled)
            self._commit_scheduled = None

        self._commit()
        self._con.close()
        del self._con



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import typing

import logging
import sys
from pathlib import Path

from gajim.common import app
from gajim.common import configpaths

from gajim.gtk.audio_player import AudioPlayer

if sys.platform == "win32" or typing.TYPE_CHECKING:
    import winsound

if sys.platform == "darwin" or typing.TYPE_CHECKING:
    from AppKit import NSSound

log = logging.getLogger("gajim.c.sound")


class PlaySound:
    def play(self, path: Path, loop: bool = False) -> None:
        raise NotImplementedError

    def stop(self) -> None:
        raise NotImplementedError

    def loop_in_progress(self) -> bool:
        raise NotImplementedError


class PlatformWindows(PlaySound):
    def __init__(self) -> None:
        self._loop_in_progress = False

    def play(self, path: Path, loop: bool = False) -> None:
        if self._loop_in_progress:
            return

        assert winsound is not None
        flags = winsound.SND_FILENAME | winsound.SND_ASYNC | winsound.SND_NODEFAULT
        if loop:
            self._loop_in_progress = True
            flags = flags | winsound.SND_LOOP

        try:
            winsound.PlaySound(str(path), flags)
        except Exception:
            log.exception("Sound Playback Error")

    def stop(self) -> None:
        assert winsound is not None
        try:
            winsound.PlaySound(None, 0)
        except Exception:
            log.exception("Sound Playback Error")
        self._loop_in_progress = False

    def loop_in_progress(self) -> bool:
        return self._loop_in_progress


class PlatformMacOS(PlaySound):
    def play(self, path: Path, loop: bool = False) -> None:
        assert NSSound is not None
        sound = NSSound.alloc()
        sound.initWithContentsOfFile_byReference_(str(path), True)
        sound.play()

    def stop(self) -> None:
        pass

    def loop_in_progress(self) -> bool:
        return False


class PlatformUnix(PlaySound):
    _SOUND_PREVIEW_ID = 0

    def __init__(self) -> None:
        self._loop_in_progress = False
        self._audio_player = AudioPlayer()

    @property
    def sound_preview_id(self) -> int:
        return self._SOUND_PREVIEW_ID

    def play(self, path: Path, loop: bool = False) -> None:
        if self._loop_in_progress:
            return

        self.stop()
        self._audio_player.get_audio_state(self._SOUND_PREVIEW_ID)
        self._loop_in_progress = loop
        self._audio_player.play_audio_file(
            path, self._SOUND_PREVIEW_ID, loop=loop, from_start=True
        )

    def stop(self) -> None:
        self._loop_in_progress = False
        self._audio_player.stop(self._SOUND_PREVIEW_ID)

    def loop_in_progress(self) -> bool:
        return self._loop_in_progress


def check_soundfile_path(file_: str, dirs: list[Path] | None = None) -> Path | None:
    """
    Check if the sound file exists

    :param file_: the file to check, absolute or relative to 'dirs' path
    :param dirs: list of knows paths to fallback if the file doesn't exists
                                     (eg: ~/.gajim/sounds/, DATADIR/sounds...).
    :return      the path to file or None if it doesn't exists.
    """
    if not file_:
        return None
    if Path(file_).exists():
        return Path(file_)

    if dirs is None:
        dirs = [configpaths.get("MY_DATA"), configpaths.get("DATA")]

    for dir_ in dirs:
        dir_ = dir_ / "sounds" / file_
        if dir_.exists():
            return dir_
    return None


def allow_sound_notification(account: str | None, sound_event: str) -> bool:
    if not app.settings.get("sounds_on"):
        return False

    if account is None:
        return True

    client = app.get_client(account)
    if client.status != "online" and not app.settings.get("sounddnd"):
        return False
    if app.settings.get_soundevent_settings(sound_event)["enabled"]:  # noqa: SIM103
        return True
    return False


def play(
    sound_event: str,
    account: str | None,
    *,
    force: bool = False,
    loop: bool = False,
) -> None:

    if force or allow_sound_notification(account, sound_event):
        str_path_to_soundfile = typing.cast(
            str, app.settings.get_soundevent_settings(sound_event)["path"]
        )
        path_to_soundfile = check_soundfile_path(str_path_to_soundfile)
        if path_to_soundfile is None:
            return

        _platform_player.play(path_to_soundfile, loop)


def stop():
    _platform_player.stop()


def _init_platform() -> PlaySound:
    if sys.platform == "win32":
        return PlatformWindows()

    if sys.platform == "darwin":
        return PlatformMacOS()

    return PlatformUnix()


_platform_player = _init_platform()



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from gi.repository import Adw
from gi.repository import Gdk
from gi.repository import GLib
from gi.repository import GObject
from gi.repository import Gtk

from gajim.common import app
from gajim.common.i18n import _
from gajim.common.open_graph_parser import OpenGraphData
from gajim.common.util.image import get_texture_from_data
from gajim.common.util.text import to_one_line

from gajim.gtk.menus import get_preview_menu
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import check_finalize
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.widgets import GajimPopover


@Gtk.Template.from_string(string=get_ui_string("preview/open_graph.ui"))
class OpenGraphPreviewWidget(Gtk.Box, SignalManager):
    __gtype_name__ = "OpenGraphPreviewWidget"
    __gsignals__ = {
        "remove": (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    _content_box: Gtk.Box = Gtk.Template.Child()
    _placeholder_image: Gtk.Image = Gtk.Template.Child()
    _picture_clamp: Adw.Clamp = Gtk.Template.Child()
    _picture: Gtk.Picture = Gtk.Template.Child()
    _loading_spinner: Adw.Spinner = Gtk.Template.Child()
    _title_label: Gtk.Label = Gtk.Template.Child()
    _description_label: Gtk.Label = Gtk.Template.Child()
    _close_button: Gtk.Button = Gtk.Template.Child()

    def __init__(
        self,
        about_url: str,
        *,
        og_data: OpenGraphData | None = None,
        pk: int | None = None,
        minimal: bool = False,
    ) -> None:
        Gtk.Box.__init__(self)
        SignalManager.__init__(self)
        self.set_cursor(Gdk.Cursor.new_from_name("pointer"))

        self._about_url = about_url
        self._og_data = og_data

        self.set_open_graph(og_data, minimal=minimal)
        self.set_tooltip_text(_("Open %s") % self._about_url)

        self._menu_popover = GajimPopover(None)
        self.append(self._menu_popover)

        gesture_primary_click = Gtk.GestureClick(button=Gdk.BUTTON_PRIMARY)
        self._connect(gesture_primary_click, "pressed", self._on_primary_clicked)
        self._content_box.add_controller(gesture_primary_click)

        gesture_secondary_click = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        self._connect(gesture_secondary_click, "pressed", self._on_secondary_clicked)
        self._content_box.add_controller(gesture_secondary_click)

        self._connect(self._close_button, "clicked", self._on_close_clicked)
        self._pk = pk

    def set_open_graph(self, og_data: OpenGraphData | None, *, minimal: bool) -> None:
        self._og_data = og_data

        if og_data is None:
            self._loading_spinner.set_visible(True)
            return

        self._loading_spinner.set_visible(False)
        self._title_label.set_text(og_data.title or "")

        if minimal:
            return

        if thumbnail := og_data.thumbnail:
            texture = get_texture_from_data(thumbnail.data)
            self._picture.set_paintable(texture)
            self._picture_clamp.set_visible(True)
            self._placeholder_image.set_visible(False)

        if description := og_data.description:
            if len(description) > 100:
                self._description_label.set_text(f"{to_one_line(description)[:100]}…")
                self._description_label.set_tooltip_text(description)
            else:
                self._description_label.set_text(to_one_line(description))
            self._description_label.set_visible(True)

    def get_open_graph(self) -> OpenGraphData | None:
        return self._og_data

    def set_error(self) -> None:
        self._loading_spinner.set_visible(False)
        self._title_label.set_text(_("Could not generate link preview"))

    def _on_primary_clicked(
        self,
        _gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> int:
        app.app.activate_action("open-link", GLib.Variant("s", self._about_url))
        return Gdk.EVENT_STOP

    def _on_secondary_clicked(
        self,
        _gesture_click: Gtk.GestureClick,
        _n_press: int,
        x: float,
        y: float,
    ) -> None:
        assert self._about_url is not None
        menu = get_preview_menu(self._about_url)
        self._menu_popover.set_menu_model(menu)
        self._menu_popover.set_pointing_to_coord(x, y)
        self._menu_popover.popup()

    def _on_close_clicked(self, _button: Gtk.Button) -> None:
        if self._pk is not None:
            app.storage.archive.remove_og(self._pk)
            parent = self.get_parent()
            assert isinstance(parent, Gtk.Box)
            parent.remove(self)

        self.emit("remove")
        check_finalize(self)

    def do_unroot(self) -> None:
        del self._menu_popover
        self._disconnect_all()
        Gtk.Box.do_unroot(self)



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

from typing import Any

from gi.repository import GLib
from gi.repository import Gtk

from gajim.common import app
from gajim.common import ged
from gajim.common import types
from gajim.common.const import AvatarSize
from gajim.common.const import SimpleClientState
from gajim.common.const import XmppUriQuery
from gajim.common.events import AccountEnabled
from gajim.common.events import BookmarksReceived
from gajim.common.events import MessageReceived
from gajim.common.ged import EventHelper
from gajim.common.helpers import generate_qr_code
from gajim.common.i18n import _
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.contacts import GroupchatContact
from gajim.common.modules.contacts import GroupchatParticipant
from gajim.common.storage.archive.const import ChatDirection
from gajim.common.storage.archive.const import MessageType
from gajim.common.util.text import make_href_markup

from gajim.gtk.contact_popover import ContactPopover
from gajim.gtk.groupchat_voice_requests_button import VoiceRequestsButton
from gajim.gtk.menus import get_banner_narrow_menu
from gajim.gtk.menus import get_groupchat_menu
from gajim.gtk.menus import get_private_chat_menu
from gajim.gtk.menus import get_self_contact_menu
from gajim.gtk.menus import get_singlechat_menu
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import get_ui_string
from gajim.gtk.widgets import AccountBadge


@Gtk.Template(string=get_ui_string("chat_banner.ui"))
class ChatBanner(Gtk.Box, EventHelper, SignalManager):
    __gtype_name__ = "ChatBanner"

    _share_popover: Gtk.Popover = Gtk.Template.Child()
    _share_instructions: Gtk.Label = Gtk.Template.Child()
    _qr_code_image: Gtk.Image = Gtk.Template.Child()
    _jid_label: Gtk.Label = Gtk.Template.Child()
    _copy_jid_button: Gtk.Button = Gtk.Template.Child()
    _avatar_button: Gtk.MenuButton = Gtk.Template.Child()
    _avatar_image: Gtk.Image = Gtk.Template.Child()
    _name_label: Gtk.Label = Gtk.Template.Child()
    _phone_image: Gtk.Image = Gtk.Template.Child()
    _robot_image: Gtk.Image = Gtk.Template.Child()
    _description_label: Gtk.Label = Gtk.Template.Child()
    _additional_items_box: Gtk.Box = Gtk.Template.Child()
    _share_menu_button: Gtk.MenuButton = Gtk.Template.Child()
    _contact_info_button: Gtk.Button = Gtk.Template.Child()
    _muc_invite_button: Gtk.Button = Gtk.Template.Child()
    _toggle_roster_button: Gtk.Button = Gtk.Template.Child()
    _toggle_roster_image: Gtk.Image = Gtk.Template.Child()
    _chat_menu_button: Gtk.MenuButton = Gtk.Template.Child()

    def __init__(self) -> None:
        Gtk.Box.__init__(self)
        EventHelper.__init__(self)
        SignalManager.__init__(self)
        self.set_name("ChatBanner")

        self._client: types.Client | None = None
        self._contact: types.ChatContactT | None = None
        self._narrow = False

        self._last_message_from_phone: set[BareContact] = set()

        self._account_badge = AccountBadge(bind_setting=True)
        self._account_badge.set_valign(Gtk.Align.CENTER)
        self._additional_items_box.append(self._account_badge)

        self._voice_requests_button = VoiceRequestsButton()
        self._additional_items_box.append(self._voice_requests_button)

        self._avatar_button.set_create_popup_func(self._on_avatar_clicked)

        hide_roster = app.settings.get("hide_groupchat_occupants_list")
        self._set_toggle_roster_button_icon(hide_roster)

        self._connect(self._copy_jid_button, "clicked", self._on_copy_jid_clicked)
        self._connect(
            self._toggle_roster_button, "clicked", self._on_toggle_roster_clicked
        )
        self._connect(
            self._share_menu_button, "notify::active", self._on_share_activated
        )
        self._chat_menu_button.set_create_popup_func(self._set_chat_menu)

        app.settings.connect_signal(
            "hide_groupchat_occupants_list", self._set_toggle_roster_button_icon
        )

    def clear(self) -> None:
        self._disconnect_signals()
        self.unregister_events()

    def switch_contact(self, contact: types.ChatContactT) -> None:
        self._disconnect_signals()
        self._client = app.get_client(contact.account)
        self._contact = contact
        self._connect_signals()

        if not self.has_events_registered():
            self.register_events(
                [
                    ("message-received", ged.GUI2, self._on_message_received),
                    ("bookmarks-received", ged.GUI2, self._on_bookmarks_received),
                    ("account-enabled", ged.GUI2, self._on_account_changed),
                    ("account-disabled", ged.GUI2, self._on_account_changed),
                ]
            )

        self._voice_requests_button.switch_contact(self._contact)

        self._update_phone_image()
        self._update_robot_image()
        self._update_roster_button()
        self._update_invite_button()
        self._update_contact_info_button()
        self._update_avatar()
        self._update_avatar_button_sensitivity()
        self._update_name_label()
        self._update_description_label()
        self._update_account_badge()
        self._update_share_box()

    def _connect_signals(self) -> None:
        assert self._contact is not None
        assert self._client is not None

        self._contact.multi_connect(
            {
                "nickname-update": self._on_name_update,
                "avatar-update": self._on_avatar_update,
                "presence-update": self._on_presence_update,
                "caps-update": self._on_caps_update,
            }
        )

        if isinstance(self._contact, GroupchatContact):
            self._contact.multi_connect(
                {
                    "room-voice-request": self._on_room_voice_request,
                    "disco-info-update": self._on_disco_info_update,
                    "name-update": self._on_name_update,
                }
            )

        elif isinstance(self._contact, GroupchatParticipant):
            self._contact.multi_connect(
                {
                    "user-joined": self._on_user_state_changed,
                    "user-left": self._on_user_state_changed,
                    "user-avatar-update": self._on_user_avatar_update,
                }
            )

        self._client.connect_signal("state-changed", self._on_client_state_changed)

    def _disconnect_signals(self) -> None:
        if self._contact is not None:
            self._contact.disconnect_all_from_obj(self)
            self._contact = None

        if self._client is not None:
            self._client.disconnect_all_from_obj(self)
            self._client = None

    def _on_client_state_changed(
        self, _client: types.Client, _signal_name: str, state: SimpleClientState
    ):
        self._update_avatar()

    def _on_presence_update(
        self, _contact: types.BareContact, _signal_name: str
    ) -> None:
        self._update_avatar()
        self._update_description_label()

    def _on_name_update(self, _contact: types.ChatContactT, _signal_name: str) -> None:
        self._update_name_label()

    def _on_avatar_update(self, _contact: types.BareContact, _signal_name: str) -> None:
        self._update_avatar()

    def _on_caps_update(self, _contact: types.BareContact, _signal_name: str) -> None:
        self._update_avatar()
        self._update_robot_image()

    def _on_room_voice_request(self, *args: Any) -> None:
        self._voice_requests_button.set_visible(True)

    def _on_user_state_changed(self, *args: Any) -> None:
        self._update_avatar()

    def _on_user_avatar_update(self, *args: Any) -> None:
        self._update_avatar()

    def _on_bookmarks_received(self, _event: BookmarksReceived) -> None:
        if not isinstance(self._contact, GroupchatContact):
            return

        self._update_name_label()

    def _on_disco_info_update(
        self, _contact: GroupchatContact, _signal_name: str
    ) -> None:
        self._update_name_label()
        self._update_description_label()

    def _on_account_changed(self, event: AccountEnabled) -> None:
        self._update_account_badge()

    def _on_message_received(self, event: MessageReceived) -> None:
        assert self._contact is not None
        if event.jid != self._contact.jid:
            return

        if event.from_mam or event.m_type != MessageType.CHAT:
            return

        message = event.message
        if message.direction == ChatDirection.OUTGOING:
            return

        if message.resource is None:
            return

        assert isinstance(self._contact, BareContact)

        resource_contact = self._contact.get_resource(message.resource)
        if resource_contact.is_phone:
            self._last_message_from_phone.add(self._contact)
        else:
            self._last_message_from_phone.discard(self._contact)

        self._update_phone_image()

    def set_narrow(self, narrow: bool) -> None:
        if self._narrow == narrow:
            return
        self._narrow = narrow

        # In narrow mode contact-info/share/invite buttons are
        # folded into the 3-dot menu.
        if narrow:
            self._share_menu_button.set_popover(None)
            self._share_popover.set_parent(self._chat_menu_button)
        else:
            self._share_popover.unparent()
            self._share_menu_button.set_popover(self._share_popover)

        if self._contact is not None:
            self._update_invite_button()
            self._update_share_box()
            self._update_contact_info_button()

    def present_share_popover(self) -> None:
        self._update_share_popover()
        self._share_popover.popup()

    def _set_chat_menu(self, menu_button: Gtk.MenuButton) -> None:
        assert self._contact is not None

        if isinstance(self._contact, GroupchatContact):
            menu = get_groupchat_menu(self._contact)

        elif isinstance(self._contact, GroupchatParticipant):
            menu = get_private_chat_menu(self._contact)

        elif self._contact.is_self:
            menu = get_self_contact_menu(self._contact)

        else:
            menu = get_singlechat_menu(self._contact)

        if self._narrow:
            menu = get_banner_narrow_menu(
                self._contact.account, self._contact.jid, menu
            )

        menu_button.set_menu_model(menu)

    def _update_phone_image(self) -> None:
        self._phone_image.set_visible(self._contact in self._last_message_from_phone)

    def _update_robot_image(self) -> None:
        if isinstance(self._contact, BareContact):
            self._robot_image.set_visible(
                self._contact.is_gateway or self._contact.is_bot
            )
        else:
            self._robot_image.set_visible(False)

    def _update_roster_button(self) -> None:
        self._toggle_roster_button.set_visible(
            isinstance(self._contact, GroupchatContact)
        )

    def _update_invite_button(self) -> None:
        self._muc_invite_button.set_visible(
            isinstance(self._contact, GroupchatContact) and not self._narrow
        )

    def _update_contact_info_button(self) -> None:
        self._contact_info_button.set_visible(not self._narrow)

    def _update_avatar(self) -> None:
        scale = app.window.get_scale_factor()
        assert self._contact
        texture = self._contact.get_avatar(AvatarSize.CHAT_BANNER, scale)
        self._avatar_image.set_pixel_size(AvatarSize.CHAT_BANNER)
        self._avatar_image.set_from_paintable(texture)

    def _update_avatar_button_sensitivity(self) -> None:
        self._avatar_button.set_sensitive(isinstance(self._contact, BareContact))

    def _on_avatar_clicked(self, button: Gtk.MenuButton) -> None:
        if not isinstance(self._contact, BareContact):
            button.set_popover(None)
            return

        button.set_popover(ContactPopover(self._contact))

    def _update_name_label(self) -> None:
        assert self._contact is not None

        name = self._get_name_from_contact(self._contact)
        self._name_label.set_markup(f"<span>{GLib.markup_escape_text(name)}</span>")
        self._name_label.set_tooltip_text(name)

    def _get_muc_description_text(self) -> str | None:
        contact = self._contact
        assert isinstance(contact, GroupchatContact)

        disco_info = app.storage.cache.get_last_disco_info(contact.jid)
        if disco_info is None:
            return None

        return disco_info.muc_description

    def _update_description_label(self) -> None:
        contact = self._contact
        assert contact is not None

        if contact.is_groupchat:
            text = self._get_muc_description_text()
        else:
            assert not isinstance(contact, GroupchatContact)
            text = contact.status or ""
        self._description_label.set_markup(make_href_markup(text))
        self._description_label.set_visible(bool(text))

    def _update_account_badge(self) -> None:
        if self._contact is None:
            self._account_badge.set_visible(False)
            return

        visible = len(app.settings.get_active_accounts()) > 1
        if visible:
            self._account_badge.set_account(self._contact.account)

        self._account_badge.set_visible(visible)

    def _update_share_box(self) -> None:
        assert self._contact is not None
        if self._contact.is_groupchat:
            self._share_menu_button.set_tooltip_text(_("Share Group Chat…"))
        else:
            self._share_menu_button.set_tooltip_text(_("Share Contact…"))
        self._share_menu_button.set_sensitive(not self._contact.is_pm_contact)
        self._share_menu_button.set_visible(not self._narrow)
        self._jid_label.set_text(str(self._contact.jid))

    def _get_share_uri(self) -> str:
        assert self._client is not None
        assert self._contact is not None
        jid = self._contact.get_address()
        if self._contact.is_groupchat:
            return jid.to_iri(XmppUriQuery.JOIN.value)
        else:
            return self._client.get_module("OMEMO").compose_trust_uri(jid)

    def _on_share_activated(self, _button: Gtk.MenuButton, *args: Any) -> None:
        self._update_share_popover()

    def _update_share_popover(self) -> None:
        assert self._contact is not None
        if isinstance(self._contact, GroupchatContact):
            share_text = _("Scan this QR code to join %s.")
            if self._contact.muc_context == "private":
                share_text = _("%s can be joined by invite only.")
        else:
            share_text = _("Scan this QR code to start a chat with %s.")
        self._share_instructions.set_text(share_text % self._contact.name)

        if (
            isinstance(self._contact, GroupchatContact)
            and self._contact.muc_context == "private"
        ):
            # Don't display QR code for private MUCs (they require an invite)
            self._qr_code_image.set_visible(False)
            return

        # Generate QR code on demand (i.e. not when switching chats)
        self._qr_code_image.set_from_paintable(generate_qr_code(self._get_share_uri()))
        self._qr_code_image.set_visible(True)

    def _on_copy_jid_clicked(self, _button: Gtk.Button) -> None:
        self.get_clipboard().set(self._get_share_uri())
        self._share_popover.popdown()

    def _on_toggle_roster_clicked(self, _button: Gtk.Button) -> None:
        state = app.settings.get("hide_groupchat_occupants_list")
        app.settings.set("hide_groupchat_occupants_list", not state)

    def _set_toggle_roster_button_icon(self, hide_roster: bool, *args: Any) -> None:
        icon = (
            "lucide-chevron-right-symbolic"
            if not hide_roster
            else "lucide-chevron-left-symbolic"
        )
        self._toggle_roster_image.set_from_icon_name(icon)

    @staticmethod
    def _get_name_from_contact(contact: types.ChatContactT) -> str:
        name = contact.name

        if isinstance(contact, BareContact):
            if contact.is_self:
                return _("Note to myself")
            return name

        if isinstance(contact, GroupchatParticipant):
            return f"{name} ({contact.room.name})"

        return name



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

# XEP-0172: User Nickname

from __future__ import annotations

from typing import Any

from nbxmpp.namespaces import Namespace
from nbxmpp.structs import MessageProperties
from nbxmpp.structs import StanzaHandler

from gajim.common import app
from gajim.common import types
from gajim.common.modules.base import BaseModule
from gajim.common.modules.contacts import BareContact
from gajim.common.modules.util import event_node


class UserNickname(BaseModule):
    _nbxmpp_extends = "Nickname"
    _nbxmpp_methods = [
        "set_nickname",
        "set_access_model",
    ]

    def __init__(self, con: types.Client):
        BaseModule.__init__(self, con)

        self.handlers = [
            StanzaHandler(
                name="message",
                callback=self._message_nickname_received,
                typ="chat",
                priority=51,
            ),
        ]

        self._register_pubsub_handler(self._nickname_received)

    def _message_nickname_received(
        self, _con: types.NBXMPPClient, _stanza: Any, properties: MessageProperties
    ) -> None:

        if properties.nickname is None:
            return

        remote_jid = properties.remote_jid
        assert remote_jid is not None

        contact = self._client.get_module("Contacts").get_contact(remote_jid)
        if not isinstance(contact, BareContact):
            return

        if contact.subscription in ("both", "to"):
            # When we are subscripted to this contact we can
            # get the nickname via pubsub
            return

        if contact.name == properties.nickname:
            return

        app.storage.archive.set_contact_value(
            self._account,
            remote_jid,
            "remote_name",
            f"{properties.nickname} ({remote_jid})",
        )

        contact.notify("nickname-update")

    @event_node(Namespace.NICK)
    def _nickname_received(
        self, _con: types.NBXMPPClient, _stanza: Any, properties: MessageProperties
    ) -> None:
        assert properties.pubsub_event is not None
        if properties.pubsub_event.retracted:
            return

        nick = properties.pubsub_event.data
        assert properties.jid is not None

        app.storage.archive.set_contact_value(
            self._account,
            properties.jid,
            "remote_name",
            nick,
        )

        self._log.info("Nickname for %s: %s", properties.jid, nick)

        contact = self._con.get_module("Contacts").get_contact(properties.jid)
        contact.notify("nickname-update")



# This file is part of Gajim.
#
# SPDX-License-Identifier: GPL-3.0-only

from __future__ import annotations

import logging

import emoji
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Pango

from gajim.common.styling import BaseHyperlink
from gajim.common.styling import PlainBlock
from gajim.common.styling import process_uris
from gajim.common.util.text import process_non_spacing_marks
from gajim.common.util.uri import parse_uri

from gajim.gtk.const import MAX_MESSAGE_LENGTH
from gajim.gtk.menus import get_conv_action_context_menu
from gajim.gtk.menus import get_uri_context_menu
from gajim.gtk.util.classes import SignalManager
from gajim.gtk.util.misc import make_pango_attributes
from gajim.gtk.util.misc import open_uri

log = logging.getLogger("gajim.gtk.conversaion.plain_widget")

URI_TAGS = ["uri", "address", "xmppadr", "mailadr"]
STYLE_TAGS = ["strong", "emphasis", "strike", "pre"]


class PlainWidget(Gtk.Box):
    def __init__(self, account: str, selectable: bool) -> None:
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        self.set_vexpand(True)

        self._account = account

        self._text_widget = MessageLabel(self._account, selectable)
        self.append(self._text_widget)

    def set_selectable(self, selectable: bool) -> None:
        self._text_widget.set_selectable(selectable)

    def add_content(self, block: PlainBlock) -> None:
        self._text_widget.print_text_with_styling(block)

    def add_action_phrase(self, text: str, nickname: str) -> None:
        self._text_widget.add_action_phrase(text, nickname)


class MessageLabel(Gtk.Label, SignalManager):
    def __init__(self, account: str, selectable: bool) -> None:
        Gtk.Label.__init__(
            self,
            hexpand=True,
            selectable=selectable,
            xalign=0,
            wrap=True,
            wrap_mode=Pango.WrapMode.WORD_CHAR,
        )
        SignalManager.__init__(self)

        self._account = account

        self.add_css_class("document")
        self.add_css_class("gajim-conversation-text")

        self._connect(self, "activate-link", self._on_activate_link)

        gesture_secondary_click = Gtk.GestureClick(button=Gdk.BUTTON_SECONDARY)
        self._connect(gesture_secondary_click, "pressed", self._on_secondary_clicked)
        self.add_controller(gesture_secondary_click)

        focus_controller = Gtk.EventControllerFocus()
        self._connect(focus_controller, "enter", self._on_focus_enter)
        self._connect(focus_controller, "leave", self._on_focus_leave)
        self.add_controller(focus_controller)

    def do_unroot(self) -> None:
        self._disconnect_all()
        Gtk.Label.do_unroot(self)

    def _on_secondary_clicked(
        self,
        gesture_click: Gtk.GestureClick,
        _n_press: int,
        _x: float,
        _y: float,
    ) -> None:
        current_uri = self.get_current_uri()
        selected, start, end = self.get_selection_bounds()

        uri_menu = None
        action_menu = None
        if current_uri:
            uri_menu = get_uri_context_menu(self._account, parse_uri(current_uri))
        elif selected:
            selected_text = self.get_text()[start:end]
            action_menu = get_conv_action_context_menu(self._account, selected_text)

        if uri_menu is None and action_menu is None:
            return

        extra_menu = Gio.Menu()
        if uri_menu is not None:
            extra_menu.append_section(None, uri_menu)
        if action_menu is not None:
            extra_menu.append_section(None, action_menu)

        self.set_extra_menu(extra_menu)
        return

    def _build_link_markup(self, text: str, uris: list[BaseHyperlink]) -> str:
        markup_text = ""
        after = GLib.markup_escape_text(text.strip())
        for uri in uris:
            uri_escaped = GLib.markup_escape_text(uri.text)
            before, _, after = after.partition(uri_escaped)
            markup_text += before
            markup_text += uri.get_markup_string()
        markup_text += after
        return markup_text

    def print_text_with_styling(self, block: PlainBlock) -> None:
        text = self._build_link_markup(block.text, block.uris)
        self.set_markup(process_non_spacing_marks(text))
        if len(self.get_text()) > MAX_MESSAGE_LENGTH:
            # Limit message styling processing
            return

        self.set_attributes(make_pango_attributes(block))

        stripped = text.strip()
        # https://github.com/carpedm20/emoji/issues/300
        # purely_emoji() return True for the empty string
        if stripped and emoji.purely_emoji(stripped):
            emoji_count = emoji.emoji_count(stripped)
            if emoji_count == 1:
                classname = "gajim-single-emoji-msg"
            else:
                classname = "gajim-emoji-msg"
            self.add_css_class(classname)

    def add_action_phrase(self, text: str, nickname: str) -> None:
        text = text.replace("/me", f"* {nickname}", 1)
        uris = process_uris(text)
        text = self._build_link_markup(text, uris)
        self.set_markup(process_non_spacing_marks(f"<i>{text}</i>"))

    def _on_activate_link(self, _label: Gtk.Label, uri: str) -> int:
        puri = parse_uri(uri)
        open_uri(puri)
        return Gdk.EVENT_STOP

    def _on_focus_enter(self, _focus_controller: Gtk.EventControllerFocus) -> None:
        self.remove_css_class("transparent-selection")

    def _on_focus_leave(self, _focus_controller: Gtk.EventControllerFocus) -> None:
        self.add_css_class("transparent-selection")
